Package: leafgl
Title: High-Performance 'WebGl' Rendering for Package 'leaflet'
Version: 0.2.4
Authors@R: 
    c(person(given = "Tim",
             family = "Appelhans",
             role = c("cre", "aut", "cph"),
             email = "tappelhans@tutamail.com",
             comment = c(ORCID = "0000-0002-9824-2707")),
      person(given = "Colin",
             family = "Fay",
             role = "ctb",
             email = "colin@thinkr.fr",
             comment = c(ORCID = "0000-0001-7343-1846")),
      person(given = "Robert",
             family = "Plummer",
             role = "ctb",
             comment = "Leaflet.glify plugin"),
      person(given = "Kent",
             family = "Johnson",
             role = "ctb",
             email = "kjohnson@akoyabio.com"),
      person(given = "Sebastian",
             family = "Gatscha",
             role = "ctb",
             email = "sebastian_gatscha@gmx.at"),
      person(given = "Olivier", 
             family = "Roy", 
             role = "ctb"))
Description: Provides bindings to the 'Leaflet.glify' JavaScript library which extends the 'leaflet' JavaScript library to render large data in the browser using 'WebGl'.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Imports: htmltools, leaflet, sf, yyjsonr, grDevices
Suggests: colourvalues, sp, shiny, testthat (>= 2.1.0)
URL: https://github.com/r-spatial/leafgl,
        https://r-spatial.github.io/leafgl/
BugReports: https://github.com/r-spatial/leafgl/issues
NeedsCompilation: no
Packaged: 2026-04-15 08:22:08 UTC; tim
Author: Tim Appelhans [cre, aut, cph] (ORCID:
    <https://orcid.org/0000-0002-9824-2707>),
  Colin Fay [ctb] (ORCID: <https://orcid.org/0000-0001-7343-1846>),
  Robert Plummer [ctb] (Leaflet.glify plugin),
  Kent Johnson [ctb],
  Sebastian Gatscha [ctb],
  Olivier Roy [ctb]
Maintainer: Tim Appelhans <tappelhans@tutamail.com>
Repository: CRAN
Date/Publication: 2026-04-15 09:10:02 UTC
Built: R 4.6.0; ; 2026-04-28 23:34:38 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(makeColorMatrix,Date)
S3method(makeColorMatrix,POSIXct)
S3method(makeColorMatrix,POSIXlt)
S3method(makeColorMatrix,character)
S3method(makeColorMatrix,data.frame)
S3method(makeColorMatrix,factor)
S3method(makeColorMatrix,formula)
S3method(makeColorMatrix,integer)
S3method(makeColorMatrix,json)
S3method(makeColorMatrix,list)
S3method(makeColorMatrix,matrix)
S3method(makeColorMatrix,numeric)
S3method(makePopup,Spatial)
S3method(makePopup,character)
S3method(makePopup,data.frame)
S3method(makePopup,default)
S3method(makePopup,formula)
S3method(makePopup,json)
S3method(makePopup,list)
S3method(makePopup,logical)
S3method(makePopup,matrix)
S3method(makePopup,sf)
S3method(makePopup,shiny.tag)
export(addGlPoints)
export(addGlPolygons)
export(addGlPolylines)
export(clearGlGroup)
export(clearGlLayers)
export(leafglOutput)
export(makeColorMatrix)
export(makePopup)
export(removeGlPoints)
export(removeGlPolygons)
export(removeGlPolylines)
export(renderLeafgl)
importFrom(htmltools,htmlDependencies)
importFrom(htmltools,tagList)
importFrom(htmltools,tags)
importFrom(leaflet,evalFormula)
importFrom(leaflet,leafletOutput)
importFrom(leaflet,renderLeaflet)

# leafgl - performant WebGL rendering for leaflet

[![R-CMD-check](https://github.com/r-spatial/leafgl/workflows/R-CMD-check/badge.svg)](https://github.com/r-spatial/leafgl/actions)
[![cran
checks](https://badges.cranchecks.info/worst/leafgl.svg)](https://cran.r-project.org/web/checks/check_results_leafgl.html)
![monthly](http://cranlogs.r-pkg.org/badges/leafgl)
![total](http://cranlogs.r-pkg.org/badges/grand-total/leafgl)
[![CRAN](http://www.r-pkg.org/badges/version/leafgl?color=009999)](https://cran.r-project.org/package=leafgl)

An R package for fast web gl rendering of features on leaflet maps. It’s
an R port of <https://github.com/robertleeplummerjr/Leaflet.glify> where
more detailed information/documentation can be found. Also, if you like
what you get here, make sure to star the original repo!

------------------------------------------------------------------------

### Installation

##### Stable version from CRAN

``` r
install.packages("leafgl")
```

##### Development version from Github

``` r
devtools::install_github("r-spatial/leafgl")
```

### What does it do?

It allows rendering of a large amount of features on a leaflet map. What
exactly a “large amount” is, depends

1.  on the types of features and
2.  on the hardware/software of the rendering machine.

Regarding 1. millions of points should be no problem, but millions of
polygons might, depending on their complexity - i.e. the number of
vertices.

With regard to 2., obviously the amount of RAM will matter most but
there are other, more suptle, problems that can occur.

Given it’s name, leafgl is intended to fully integrate with the leaflet
package, though it is very likely that it won’t be a 1:1 replacement for
the respective `leaflet::add*` functions. For example, given the
intention to render/visualise as many features as possible we need to
make a compromise on what additional information we allow to be part of
the rendering. So far, we allow coloring of features and popups based on
one column of the feature attributes, hence you cannot provide your own
popup content. This may seem drastic, but all this information is costly
both in terms of performance/speed and memory. In the end, who wants to
wait for a map that the browser isn’t able to render anyway…

### What doesn’t it do?

For starters, it doesn’t guarantee to be working tomorrow. At this stage
`leafgl` is pre-alpha and under heavy development so things are likely
to change frequently. For example, we are thinking about shorter funtion
names (e.g. `addGlPolygons` instead of the current `addGlifyPolygons`).
Additionally, we are still figuring out which pathway is best to use in
order to pass data from R to the browser. As a result, rendering
environments other than the browser (or RStudio viewer) may not work
properly or at all until we approach a more stable implementation.

#### NOTE

Depending on your operating system and browser, you may see some weird
colors that do not correspond to the ones that you specified. The only
known work-around at this stage is to set `opacity = 1`. For more
details the inclined reader is referred to [this
issue](https://github.com/r-spatial/leafgl/issues/4)

### What can I do to help?

A lot! First and foremost you can use it as often as possible and report
issues/bugreports and/or feature request (see end of page for details).
If you have ideas on how to enhance functionality without impacting
performance too much and feel confident enough to provide pull request,
please don’t hesitate. Finally, if you have proficient knowledge of
JavaScript and want/know how to improve the package in any way, we would
very much love to hear from you!

### Example usage

#### 1 Mio. points on a map

This will render 1 Mio. points on a standard leaflet map.

``` r
library(leaflet)
library(leafgl)
library(sf)

n = 1e6

df1 = data.frame(id = 1:n,
                 x = rnorm(n, 10, 3),
                 y = rnorm(n, 49, 1.8))

pts = st_as_sf(df1, coords = c("x", "y"), crs = 4326)

options(viewer = NULL) # view in browser

leaflet() %>%
  addProviderTiles(provider = providers$CartoDB.DarkMatter) %>%
  addGlPoints(data = pts, group = "pts") %>%
  setView(lng = 10.5, lat = 49.5, zoom = 6)
```

![](man/figures/pts_blue.png)

<br>

------------------------------------------------------------------------

#### Colouring points by value mapping

For this we use `library(colourvalues)` because it can create color
voctors in the blink of an eye!

``` r
library(leaflet)
library(leafgl)
library(sf)
library(colourvalues)

n = 1e6

df1 = data.frame(id = 1:n,
                 x = rnorm(n, 10, 3),
                 y = rnorm(n, 49, 1.8))

pts = st_as_sf(df1, coords = c("x", "y"), crs = 4326)

cols = colour_values_rgb(pts$id, include_alpha = FALSE) / 255

leaflet() %>%
  addProviderTiles(provider = providers$CartoDB.DarkMatter) %>%
  addGlPoints(data = pts, fillColor = cols, group = "pts") %>%
  setView(lng = 10.5, lat = 49.5, zoom = 6)
```

![](man/figures/pts_viridis.png)

<br>

------------------------------------------------------------------------

#### 100k polygons on a map

In reality, it only 97112 polygons… But who wants to be pedantic here?

This data was downloaded from
<https://download.geofabrik.de/europe/switzerland.html>

``` r
library(leaflet)
library(leafgl)
library(sf)
library(colourvalues)

ch_lu = st_read("/media/timpanse/d8346522-ef28-4d63-9bf3-19fec6e13aab/bu_lenovo/software/testing/mapview/switzerland/landuse.shp")

ch_lu = ch_lu[, c(1, 3, 4)] # don't handle NAs so far

options(viewer = NULL)

cols = colour_values_rgb(ch_lu$type, include_alpha = FALSE) / 255

leaflet() %>%
  addProviderTiles(provider = providers$CartoDB.DarkMatter) %>%
  addGlPolygons(data = ch_lu, 
                color = cols, 
                popup = "type",
                group = "pols") %>%
  setView(lng = 8.3, lat = 46.85, zoom = 9) %>% 
  addLayersControl(overlayGroups = "pols")
```

![](man/figures/polys_ch.png)

<br>

------------------------------------------------------------------------

## Shiny

Thanks to [@ColinFay](https://github.com/ColinFay) `leafgl` has
dedicated shiny functions. Given that what `leafgl` produces is a
`leaflet` map, we only need to use `leafglOutput` in our `ui` call. In
the `server` call we can simply use `renderLeaflet`. Here an example:

``` r
library(leaflet)
library(leafgl)
library(sf)
library(shiny)

n = 1e6

df1 = data.frame(id = 1:n,
                 x = rnorm(n, 10, 3),
                 y = rnorm(n, 49, 1.8))

pts = st_as_sf(df1, coords = c("x", "y"), crs = 4326)

options(viewer = NULL) # view in browser

m = leaflet() %>%
  addProviderTiles(provider = providers$CartoDB.DarkMatter) %>%
  addGlPoints(data = pts, group = "pts") %>%
  setView(lng = 10.5, lat = 49.5, zoom = 4) %>% 
  addLayersControl(overlayGroups = "pts")

ui <- fluidPage(
  leafglOutput("mymap")
)

server <- function(input, output, session) {
  output$mymap <- renderLeaflet(m)
}

shinyApp(ui, server)
```

<br>

------------------------------------------------------------------------

## Contact

Please file Pull requests, bug reports and feature requests at
<https://github.com/r-spatial/leafgl/issues>

library(shiny)
library(leaflet)
library(mapview)
library(leafgl)
options(shiny.autoreload = TRUE)

## Data ################
lines = suppressWarnings(st_cast(trails, "LINESTRING"));
lines = st_transform(lines, 4326)

base_colors <- "#bc2323"
ptsdata <- breweries
ptsdata$id <- seq.int(length(ptsdata$brewery))
opacity_hex <- sprintf("%02X", floor(255 * scales::rescale(ptsdata$id, to = c(0.1, 1))))
colors_with_opacity_pts <- paste0(base_colors, opacity_hex)

shapes <- franconia
base_colors <- "#5353c1"
shapes <- st_cast(shapes, "POLYGON")
opacity_hex <- sprintf("%02X", floor(255 * scales::rescale(shapes$SHAPE_AREA, to = c(1, 0.2))))
colors_with_opacity_poly <- paste0(base_colors, opacity_hex)

## contextmenu_Template ################
contextmenu_Template_Shape <- JS('
   function(e, feature) {
   $.contextMenu({
      selector: "#context-menu-anchor_shape",
      className: "shape-title",
      callback: function(key, options) {
        Shiny.setInputValue("contextMenuAction", key, {priority: "event"});
      },
      items: {
        edit: {name: "Edit Shape", icon: "edit"},
        delete: {name: "Delete  Shape", icon: "delete"},
        quit: {name: "Quit", icon: "quit"}
      }
    });

    console.log("e"); console.log(e)
    console.log("feature"); console.log(feature)
    var evt = $.Event("contextmenu", {
      pageX: e.originalEvent.pageX,
      pageY: e.originalEvent.pageY,
      which: 3  // indicate right-click
    });
    $("#context-menu-anchor_shape").trigger(evt);
  }')

contextmenu_Template_Line <- JS('
   function(e, feature) {
   $.contextMenu({
      selector: "#context-menu-anchor_lines",
      className: "line-title",
      callback: function(key, options) {
        Shiny.setInputValue("contextMenuAction", key, {priority: "event"});
      },
      items: {
        edit: {name: "Edit Line", icon: "edit"},
        delete: {name: "Delete Line", icon: "delete"},
        quit: {name: "Quit", icon: "quit"}
      }
    });

    console.log("e"); console.log(e)
    console.log("feature"); console.log(feature)
    var evt = $.Event("contextmenu", {
      pageX: e.originalEvent.pageX,
      pageY: e.originalEvent.pageY,
      which: 3  // indicate right-click
    });
    $("#context-menu-anchor_lines").trigger(evt);
  }')

contextmenu_Template_Points <- JS('
   function(e, feature) {
   $.contextMenu({
      selector: "#context-menu-anchor_points",
      className: "point-title",
      callback: function(key, options) {
        Shiny.setInputValue("contextMenuAction", key, {priority: "event"});
      },
      items: {
        edit: {name: "Edit Point", icon: "edit"},
        delete: {name: "Delete  Point", icon: "delete"},
        zoomin: {name: "Zoom In", icon: "edit"},
        zoomout: {name: "Zoom Out", icon: "edit"},
        quit: {name: "Quit", icon: "quit"}
      }
    });

    console.log("e"); console.log(e)
    console.log("feature"); console.log(feature)
    var evt = $.Event("contextmenu", {
      pageX: e.originalEvent.pageX,
      pageY: e.originalEvent.pageY,
      which: 3  // indicate right-click
    });
    $("#context-menu-anchor_points").trigger(evt);
  }')




## UI ################
ui <- fluidPage(
  tags$head(
    tags$link(rel = "stylesheet",
              href = "https://cdnjs.cloudflare.com/ajax/libs/jquery-contextmenu/2.7.1/jquery.contextMenu.min.css"),
    tags$script(src = "https://cdnjs.cloudflare.com/ajax/libs/jquery-contextmenu/2.7.1/jquery.contextMenu.min.js"),
    tags$script(src = "https://cdnjs.cloudflare.com/ajax/libs/jquery-contextmenu/2.7.1/jquery.ui.position.js")
    , tags$style("
    .line-title:before {
        content: 'Title for Lines';
    }
    .shape-title:before {
        content: 'Title for Shapes';
    }
    .point-title:before {
        content: 'Title for Points';
    }
    .line-title, .shape-title, .point-title {
        display: block;
        position: absolute;
        right: 0;
        background: #DDD;
        padding: 2px;
        font-size: 13px;
    }
    .line-title :first-child, .shape-title :first-child, .point-title :first-child {
        margin-top: 2px;
    }
    ")
  ),
  leafletOutput("map", height = 800)
  , tags$div(id = "context-menu-anchor_lines", style="display:none")
  , tags$div(id = "context-menu-anchor_points", style="display:none")
  , tags$div(id = "context-menu-anchor_shape", style="display:none")
)

## Server ################
server <- function(input, output, session) {
  output$map <- renderLeaflet({
    leaflet() %>%
      addProviderTiles(provider = "CartoDB") %>%
      addGlPolygons(
        data = shapes
        , color = colors_with_opacity_poly
        , opacity = 0.3
        , popup = TRUE
        , contextMenu = contextmenu_Template_Shape
        , group = "polys"
        ) %>%
      addGlPolylines(
        data = lines
        , layerId = lines$realid
        , weight = 2
        , color = "#528b3f"
        , sensitivity = 0.001
        , sensitivityHover = 0.001
        , label = ~FKN
        , contextMenu = contextmenu_Template_Line
        , group = "lines"
      ) %>%
      addGlPoints(
        data = ptsdata
        , layerId = ~id
        , fillColor = colors_with_opacity_pts
        , radius = ptsdata$id/10
        , contextMenu = contextmenu_Template_Points
        , group = "points"
      ) %>%
      addLayersControl(overlayGroups = c("points", "lines", "polys"))

  })
  observeEvent(input$contextMenuAction, {
    action <- input$contextMenuAction
    print(paste("Context menu action chosen:", action))
    # Handle the action (e.g., open modal, update UI, etc.)
  })

}
shinyApp(ui, server)

#' Transform object to rgb color matrix
#' @title makeColorMatrix
#' @param x Object representing the color. Can be of class integer, numeric, Date,
#'   POSIX*, character with color names or HEX codes, factor, matrix, data.frame,
#'   list, json or formula.
#' @param data The dataset
#' @param palette Name of a color palette. If \code{colourvalues} is installed, it is
#'   passed to \code{\link[colourvalues]{colour_values_rgb}}. To see all available
#'   palettes, please use \code{\link[colourvalues]{colour_palettes}}.
#'   If \code{colourvalues} is not installed, the palette is passed to
#'   \code{\link[leaflet]{colorNumeric}}.
#' @param ... Passed to \code{\link[colourvalues]{colour_palettes}} or
#'   \code{\link[leaflet]{colorNumeric}}.
#' @rdname makeColorMatrix
#' @export
#' @examples {
#' ## For Integer/Numeric/Factor
#' makeColorMatrix(23L)
#' makeColorMatrix(23)
#' makeColorMatrix(as.factor(23))
#'
#' ## For POSIXt / Date
#' makeColorMatrix(as.POSIXlt(Sys.time(), "America/New_York"), NULL)
#' makeColorMatrix(Sys.time(), NULL)
#' makeColorMatrix(Sys.Date(), NULL)
#'
#' ## For matrix/data.frame
#' makeColorMatrix(cbind(130,1,1), NULL)
#' makeColorMatrix(matrix(1:99, ncol = 3, byrow = TRUE), data.frame(x=c(1:33)))
#' makeColorMatrix(data.frame(matrix(1:99, ncol = 3, byrow = TRUE)), data.frame(x=c(1:33)))
#'
#' ## For characters
#' testdf <- data.frame(
#'   texts = LETTERS[1:10],
#'   vals = 1:10,
#'   vals1 = 11:20
#' )
#' makeColorMatrix("red", testdf)
#' makeColorMatrix("val", testdf)
#'
#' ## For formulaes
#' makeColorMatrix(~vals, testdf)
#' makeColorMatrix(~vals1, testdf)
#'
#' ## For JSON
#' makeColorMatrix(leafgl:::yyson_json_str(data.frame(r = 54, g = 186, b = 1)), NULL)
#'
#' ## For Lists
#' makeColorMatrix(list(1,2), data.frame(x=c(1,2)))
#' }
makeColorMatrix <-  function(x, data, palette, ...) {
  UseMethod("makeColorMatrix", x)
}

#' @export
makeColorMatrix.integer <- function(x, data = NULL, palette = "viridis", ...) {
  if (requireNamespace("colourvalues", quietly = TRUE)) {
    x <- checkDim(x, data)
    colourvalues::colour_values_rgb(x,
                                    palette = palette,
                                    include_alpha = FALSE, ...) / 255
  } else {
    if (length(x) == 1) {
      x <- grDevices::colors()[x]
      t(grDevices::col2rgb(x, alpha = FALSE)) / 255
    } else {
      x <- checkDim(x, data)
      pal <- leaflet::colorNumeric(palette, x, alpha = FALSE, ...)
      t(grDevices::col2rgb(pal(x), alpha = FALSE)) / 255
    }
  }
}
#' @export
makeColorMatrix.numeric <- makeColorMatrix.integer

#' @export
makeColorMatrix.factor <- function(x, data = NULL, palette = "viridis", ...) {
  x <- tryCatch(as.integer(as.character((x))),
                error = function(e) as.numeric((x)),
                warning = function(e) as.numeric((x)),
                finally = function(e) stop("Cannot process factor."))

  makeColorMatrix(x, data, palette, ...)
}

#' @export
makeColorMatrix.formula <- function(x, data, palette = "viridis", ...) {
  x <- leaflet::evalFormula(x, data)
  makeColorMatrix(x, data, palette, ...)
}

#' @export
makeColorMatrix.character <- function(x, data, palette = "viridis", ...) {
  ## If x is a column name, take the column values and feed method again
  if (length(x) == 1 && x %in% colnames(data)) {
    x <- data[[x]]
    return(makeColorMatrix(x, data, palette, ...))
  }
  ## Otherwise we assume its a color-name / HEX-code.
  ## If that errors, convert to integer/factor, and feed back
  x <- checkDim(x, data)
  col <- tryCatch(t(grDevices::col2rgb(x)) / 255,
           error = function(e) {
             x <- as.integer(as.factor(x))
             makeColorMatrix(x, data, palette, ...)
           }
  )
  col
}

#' @export
makeColorMatrix.matrix <- function(x, data, palette = "viridis", ...) {
  if (all(apply(x, 2, class) == "character")) {
    x <- matrix(apply(x, 2, as.numeric), ncol = 3)
  }
  x <- checkDim(x, data)
  if (any(x > 1)) {
    x / 255
  } else {
    x
  }
}
#' @export
makeColorMatrix.data.frame <- makeColorMatrix.matrix

#' @export
makeColorMatrix.list <- function(x, data = NULL, palette = "viridis", ...) {
  classes <- lapply(x, function(x) class(x)[[1]])
  if (all(classes == "numeric")) {
    x <- unlist(x)
  } else if (all(classes == "matrix")) {
    x <- do.call(rbind, x)
  } else {
    x <- as.numeric(unlist(x))
  }
  makeColorMatrix(x, data, palette, ...)
}

#' @export
makeColorMatrix.json <- function(x, data = NULL, palette = "viridis", ...) {
  x <- yyjsonr::read_json_str(x)
  makeColorMatrix(x, data, palette, ...)
}

#' @export
makeColorMatrix.Date <- function(x, data = NULL, palette = "viridis", ...) {
  x <- as.numeric(x)
  makeColorMatrix(x, data, palette, ...)
}
#' @export
makeColorMatrix.POSIXct <- makeColorMatrix.Date
#' @export
makeColorMatrix.POSIXlt <- makeColorMatrix.Date



#' checkDim
#' @description Check the length of the color vector. It must match the
#'   number of rows of the dataset.
#' @param x The color vector
#' @param data The dataset
checkDim <- function(x, data) {
  if (is.null(data)) { return(x) }
  if (inherits(data, "sfc")) nro_d = length(data) else nro_d = nrow(data)
  if (inherits(data, c("sf","sfc")) && length(grep("MULTI", sf::st_geometry_type(data))) > 0) {
    lnths = lengths(sf::st_geometry(data))
  } else {
    lnths = nro_d
  }
  if (inherits(x, "matrix") || inherits(x, "data.frame")) {
    if (nrow(x) != 1 && nro_d != nrow(x)) {
      warning("Number of rows of color matrix does not match number of data rows.\n",
              "  Just the first color is used.")
      x <- x[1,,drop = FALSE]
    }
  }
  else {
    len_x <- length(x)
    if ((len_x != nro_d)) {
      # warning("Length of color vector does not match number of data rows.\n",
      #         "  The vector is repeated to match the number of rows.")
      x <- rep(x, ceiling(nro_d / len_x))[1:nro_d]
    }
    if (any(lnths != 1) & length(lnths) == nro_d & length(x) != 1) {
      x = rep(x, times = lnths / 2)
    }
  }
  return(x)
}

context("test-leafgl-color_utils")
library(leaflet)
library(sf)
library(yyjsonr)

n = 1e2
df1 = data.frame(id = 1:n, id2 = n:1,
                 x = rnorm(n, 10, 1),
                 y = rnorm(n, 49, 0.8))
pts = st_as_sf(df1, coords = c("x", "y"), crs = 4326)

lines = suppressWarnings(st_cast(st_as_sf(atlStorms2005), "LINESTRING"));

polys <- suppressWarnings(st_cast(st_as_sf(gadmCHE), "POLYGON"))
multipolys <- st_as_sf(gadmCHE)

test_that("makeColorMatrix directly", {
  vec <- 1
  a <- makeColorMatrix(vec, NULL)
  expect_is(a, "matrix")
  expect_true(is.numeric(a))

  vec <- LETTERS[1:10]
  a <- makeColorMatrix(vec, NULL)
  expect_is(a, "matrix")
  expect_true(is.numeric(a))

  vec <- 1:10
  a <- makeColorMatrix(vec, NULL)
  expect_is(a, "matrix")
  expect_true(is.numeric(a))

})

test_that("Multi-geometry", {
  vec <- 1
  a <- makeColorMatrix(vec, multipolys)
  expect_is(a, "matrix")
  expect_true(is.numeric(a))
})

test_that("Character as color", {
  ## Character - Color Name ###################
  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = "red",
                group = "pts");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = "red",
                src = TRUE,
                group = "pts")
  expect_is(m, "leaflet")
  expect_identical(m$x$calls[[1]]$method, "addGlifyPointsSrc")
  rm(m)

  m <- leaflet() %>%
    addGlPoints(data = st_sfc(st_geometry(pts)),
                fillColor = "red",
                group = "pts");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPolylines(data = lines,
                color = "red",
                group = "lns");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPolygons(data = polys,
                   color = "red",
                   group = "lns");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  ## Character - Column Name ###################
  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = "id",
                group = "pts");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPolylines(data = lines,
                   color = "Name",
                   group = "lns");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPolygons(data = polys,
                  color = "NAME_1",
                  group = "lns");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  ## Character - HEX-Code ###################
  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = "#36ba01",
                group = "pts");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "list")
  expect_true(m$x$calls[[1]]$args[[2]][[1]] == "#36ba01")
  rm(m)

  m <- leaflet() %>%
    addGlPolylines(data = lines,
                   color = "#36ba01",
                   group = "lns");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "list")
  expect_true(m$x$calls[[1]]$args[[2]][[1]] == "#36ba01")
  rm(m)

  m <- leaflet() %>%
    addGlPolygons(data = polys,
                  color = "#36ba01",
                  group = "lns");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "list")
  expect_true(m$x$calls[[1]]$args[[2]][[1]] == "#36ba01")
  rm(m)

  ## Character - Character Column Name #############
  m <- leaflet() %>%
    addGlPolylines(data = lines,
                   color = "Name",
                   group = "lns");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  ## Character - Character Column Name with Palette #############
  m <- leaflet() %>%
    addGlPolylines(data = lines,
                   color = "MaxWind",
                   palette = "magma",
                   group = "lns");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)
})

test_that("Formula as color", {
  ## Formula ###################
  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = ~id,
                group = "pts");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = ~id,
                palette = "Greens",
                group = "pts");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = ~id,
                palette = "inferno",
                group = "pts",
                src = TRUE)
  expect_is(m, "leaflet")
  expect_null(m$x$calls[[1]]$args[[1]])
  rm(m)

  m <- leaflet() %>%
    addGlPolylines(data = lines,
                   color = ~Name,
                   group = "lns");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPolylines(data = lines,
                   color = ~Name,
                   palette = "Greens",
                   group = "lns");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPolylines(data = lines,
                   color = ~Name,
                   palette = "inferno",
                   group = "lns",
                   src = TRUE);
  expect_is(m, "leaflet")
  expect_null(m$x$calls[[1]]$args[[1]])
  rm(m)

  m <- leaflet() %>%
    addGlPolygons(data = polys,
                  color = ~NAME_1,
                  group = "lns");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPolygons(data = polys,
                  color = ~NAME_1,
                  palette = "BuPu",
                  group = "lns");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPolygons(data = polys,
                  color = ~NAME_1,
                  palette = "magma",
                  group = "lns",
                  src = TRUE)
  expect_is(m, "leaflet")
  expect_null(m$x$calls[[1]]$args[[1]])

})

test_that("Tables as color", {
  ## Matrix with nrow = 1 ###################
  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = cbind(180,1,10),
                group = "pts");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = cbind("180","1","10"),
                group = "pts");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = cbind(0.12, 0.9, 0.01),
                group = "pts");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPolylines(data = lines,
                   color = cbind(180,1,10),
                   group = "lns");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPolygons(data = polys,
                  color = cbind(180,1,10),
                  group = "lns");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)


  ## Matrix with nrow = nrow(data) ###################
  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = matrix(sample(1:255, nrow(pts)*3, TRUE), ncol = 3),
                group = "pts");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPolylines(data = lines,
                   color = matrix(sample(1:255, nrow(lines)*3, TRUE), ncol = 3),
                   group = "lns");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPolygons(data = polys,
                  color = matrix(sample(1:255, nrow(polys)*3, TRUE), ncol = 3),
                  group = "lns");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  ## data.frame with nrow = 1 ###################
  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = data.frame(cbind(180,1,10)),
                group = "pts");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = data.frame(matrix(sample(1:255, nrow(pts)*3, TRUE), ncol = 3)),
                group = "pts");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

})

test_that("Numeric as color", {
  ## Integer ###################
  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = 120L,
                group = "pts");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPolylines(data = lines,
                   color = 120L,
                   group = "lns");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPolygons(data = polys,
                  color = 120L,
                  group = "lns");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  ## Numeric ###################
  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = 30.43,
                group = "pts");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPolylines(data = lines,
                   color = 30.43,
                   group = "lns");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPolygons(data = polys,
                  color = 30.43,
                  group = "lns");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  ## Factor ###################
  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = as.factor(130),
                group = "pts");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = as.factor(c("asd")),
                group = "pts");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = as.factor(c("asd","bdc","fds")),
                group = "pts")
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)


  m <- leaflet() %>%
    addGlPolylines(data = lines,
                   fillColor = as.factor(130),
                   group = "lns");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPolygons(data = polys,
                  color = as.factor(130),
                  group = "lns");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

})

test_that("List as color", {
  ## List ###################
  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = list(1,2),
                group = "pts")
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = list(100),
                group = "pts")
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)


  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = list(matrix(sample(1:255, nrow(pts)*3, replace = T), ncol = 3)),
                group = "pts")
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = lapply(1:nrow(pts), function(x) matrix(sample(1:255, 3), ncol = 3)),
                group = "pts")
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = list(c(100,200), cbind(2,1)),
                group = "pts")
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- expect_warning(leaflet() %>%
                        addGlPoints(data = pts,
                                    fillColor = list(cbind(0.1,0.2,0.7),
                                                     cbind(2,1,0)),
                                    group = "pts"))
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

})

test_that("JSON as color", {
  ## JSON with same dimensions ###################
  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = leafgl:::yyson_json_str(list(matrix(sample(1:255, nrow(pts)*3, replace = T), ncol = 3))),
                group = "pts");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  ## JSON with 1 color ###################
  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = leafgl:::yyson_json_str(data.frame(r = 54, g = 186, b = 1)),
                group = "pts");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  ## JSON with wrong dimension - Warning ###################
  m <- expect_warning(leaflet() %>%
    addGlPoints(data = pts,
                fillColor = leafgl:::yyson_json_str(data.frame(r = c(54, 123),
                                                    g = c(1, 186),
                                                    b = c(1, 123))),
                group = "pts"))
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

})

test_that("Date/POSIX* as color", {
  ## POSIXlt ###################
  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = as.POSIXlt(Sys.time(), "America/New_York"),
                group = "pts");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  ## POSIXct ###################
  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = Sys.time(),
                group = "pts");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  ## Date ###################
  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = Sys.Date(),
                group = "pts");
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)
})

test_that("Warnings / Errors", {
  ## TODO Not showing anything
  m <- expect_warning(leaflet() %>%
    addGlPoints(data = pts,
                fillColor = matrix(33:98, ncol = 3, byrow = F),
                group = "pts"))
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- expect_warning(leaflet() %>%
    addGlPoints(data = pts,
                fillColor = data.frame(matrix(33:98, ncol = 3, byrow = F)),
                group = "pts"));
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  m <- leaflet() %>%
    addGlPoints(data = pts,
                fillColor = 1:33,
                group = "pts")
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[2]], "json")
  expect_true(validate_json_str(m$x$calls[[1]]$args[[2]]))
  rm(m)

  ## Errors + Warnings ###################
  expect_warning(expect_error(leaflet() %>%
    addGlPoints(data = pts,
                fillColor = cbind("asf","fasd", "fasd"),
                group = "pts")))

})

#' Transform object to popup
#' @title makePopup
#' @param x Object representing the popup
#' @param data The dataset
#' @rdname makePopup
#' @export
makePopup <-  function(x, data) {
  UseMethod("makePopup", x)
}

#' @export
makePopup.formula <- function(x, data) {
  leaflet::evalFormula(x, data)
}

#' @export
makePopup.character <- function(x, data) {
  if (all(x %in% names(data))) {
    if (length(x) == 1) {
      data[[x]]
    } else {
      names <- x
      x <- do.call(cbind, lapply(x, function(x) .subset2(data, x)))
      colnames(x) <- names
      makePopup(x, data)
    }
  } else {
    checkDimPop(x, data)
  }
}

#' @export
makePopup.shiny.tag <- function(x, data) {
  x <- as.character(x)
  x <- checkDimPop(x, data)
  if (length(x) == 1) {
    htmltools::HTML(x)
  } else {
    lapply(x, htmltools::HTML)
  }
}

#' @export
makePopup.matrix <- function(x, data) {
  names_x <- colnames(x)
  x <- apply(x, 1, function(x)
    paste0("<b>", names_x, "</b>: ", x,
          collapse = "<br>"))
  checkDimPop(x, data)
}
#' @export
makePopup.data.frame <- makePopup.matrix

#' @export
makePopup.sf <- function(x, data) {
  x <- data.frame(x)
  ## Remove Geometry Columns ???
  x[,"geometry"] <- NULL
  x[, "geom"] = NULL
  x[,"latitude"] <- NULL
  x[,"longitude"] <- NULL
  x[,"optional"] <- NULL
  ## Feed back to method
  makePopup(x, data)
}
#' @export
makePopup.Spatial <- makePopup.sf

#' @export
makePopup.logical <- function(x, data) {
  if (x == TRUE) {
    makePopup(data, data)
  } else {
    return(NULL)
  }
}

#' @export
makePopup.list <- function(x, data) {
  x <- do.call(cbind, x)
  makePopup(x, data)
}

#' @export
makePopup.json <- function(x, data) {
  x <- yyjsonr::read_json_str(x)
  makePopup(x, data)
}

#' @export
makePopup.default <- function(x, data) {
  x <- as.character(x)
  makePopup(x, data)
}



#' checkDim
#' @description Check the length of the popup vector. It must match the
#'   number of rows of the dataset.
#' @param x The popup vector
#' @param data The dataset
checkDimPop <- function(x, data) {
  if (is.null(data)) { return(x) }
  if (inherits(data, "sfc")) nro_d = length(data) else nro_d = nrow(data)
  len_x <- length(x)
  if (len_x != nro_d) {
    # warning("Length of popups does not match number of data rows.\n",
    #         "  The vector is repeated to match the number of rows.")
    x <- rep(x, ceiling(nro_d / len_x))[1:nro_d]
  }
  return(x)
}

context("test-leafgl-popup")

library(leaflet)
library(yyjsonr)
library(sf)

## makePopup solo #######
test_that("makePopup direct", {
  vec <- 1
  a <- makePopup(vec, NULL)
  expect_identical(a, as.character(vec))

  vec <- LETTERS[1:10]
  a <- makePopup(vec, NULL)
  expect_identical(a, as.character(vec))

  vec <- 1:10
  a <- makePopup(vec, NULL)
  expect_identical(a, as.character(vec))
})

## POINTS #################################
test_that("popup-points-character", {

  ## Column Name Single ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = "state",
                group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
  expect_identical(read_json_str(m$x$calls[[2]]$args[[3]]), breweries91$state)
  rm(m)

  sfpoints <- sf::st_as_sf(breweries91)
  m <- leaflet() %>%
    addGlPoints(data = st_sfc(st_geometry(sfpoints)),
                popup = sfpoints$state,
                group = "grp")
  expect_is(m, "leaflet")
  expect_is(m$x$calls[[1]]$args[[3]], "json")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[1]]$args[[3]]))
  rm(m)

  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = breweries91$state,
                group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
  expect_identical(read_json_str(m$x$calls[[2]]$args[[3]]), breweries91$state)
  rm(m)

  ## Column Names ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = c("state", "address"),
                group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
  # expect_identical(read_json_str(m$x$calls[[2]]$args[[3]]), breweries91$state)
  rm(m)

  ## Single Random Character ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = "Text 1",
                group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
  expect_identical(read_json_str(m$x$calls[[2]]$args[[3]]), rep("Text 1", nrow(breweries91)))
  rm(m)

  ## Multiple Random Characters - (wrong length) ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = c("Text 1", "Text 2"),
                group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
  expect_identical(read_json_str(m$x$calls[[2]]$args[[3]]),
                   rep(c("Text 1","Text 2"), nrow(breweries91))[1:nrow(breweries91)])
  rm(m)

  ## Multiple Random Character (same length) ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = rep("Text 1", nrow(breweries91)),
                group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
  expect_identical(read_json_str(m$x$calls[[2]]$args[[3]]), rep("Text 1", nrow(breweries91)))
})

test_that("popup-points-table", {
  ## Data.frame ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = as.data.frame(breweries91),
                group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))

  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = as.data.frame(breweries91),
                group = "grp",
                src = TRUE)
  expect_is(m, "leaflet")

  ## Data.frame - wrong length ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = as.data.frame(breweries91)[1:4,],
                group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))

  ## Matrix ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = as.matrix(as.data.frame(breweries91)),
                group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))

})

test_that("popup-points-spatial", {
  ## SpatialPointsDataFrame ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = breweries91,
                group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))

  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = breweries91,
                group = "grp",
                src = TRUE)
  expect_is(m, "leaflet")

  ## Simple Feature ##############
  library(sf)
  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = st_as_sf(breweries91),
                group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
})

test_that("popup-points-formula", {
  ## Formula ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = ~sprintf("<b>State</b>: %s<br>
                                 <b>Address</b>: %s<br>
                                 <b>Brauerei</b>: %s,",
                                 state, address, brewery),
                group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))

  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = ~sprintf("<b>State</b>: %s<br>
                                 <b>Address</b>: %s<br>
                                 <b>Brauerei</b>: %s,",
                                 state, address, brewery),
                group = "grp",
                src = TRUE)
  expect_is(m, "leaflet")
})

test_that("popup-points-list", {
  ## List with length 1 ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = as.list(data.frame(city="Berlin",
                                           district=5029, stringsAsFactors = F)),
                group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))

  ## List with length >1 ##############
  m <- leaflet() %>% addTiles() %>%
        addGlPoints(data = breweries91,
                    popup=as.list(data.frame(city=c("Vienna","Berlin"),
                                             district=c(1010,40302),
                                             stringsAsFactors = F)),
                    group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))


  ## List with same length ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = as.list(as.data.frame(breweries91)),
                group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))

})

test_that("popup-points-json", {
  ## JSON - 1 column ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = yyson_json_str(as.list(as.data.frame(breweries91[,1]))),
                group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))

  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = yyson_json_str(as.data.frame(breweries91[,1])),
                group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))

  ## JSON - multiple columns ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = yyson_json_str(as.list(as.data.frame(breweries91[,1:5]))),
                group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))

  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = yyson_json_str(as.data.frame(breweries91[1,])),
                group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))

  ## JSON with same length ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = yyson_json_str(as.list(as.data.frame(breweries91))),
                group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
})

test_that("popup-points-logical", {
  ## TRUE ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = TRUE,
                group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))

  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = TRUE,
                group = "grp",
                src = TRUE)
  expect_is(m, "leaflet")

  ## FALSE #################
  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = FALSE,
                group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
  expect_true(m$x$calls[[2]]$args[[3]] == "null")

  ## NULL #################
  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = NULL,
                group = "grp")
  expect_is(m, "leaflet")
  expect_true(is.null(m$x$calls[[2]]$args[[3]]))

})

test_that("popup-points-shiny.tag", {
  library(shiny)
  ## Shiny.Tag - icon - Length 1 ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = shiny::icon("car"),
                group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))

  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = shiny::icon("car"),
                group = "grp", src = TRUE)
  expect_is(m, "leaflet")

  ## Shiny.Tag - icon - Length >1 ##############
  # m <- expect_warning(leaflet() %>% addTiles() %>%
  #                       addGlPoints(data = breweries91,
  #                                   popup = c(
  #                                     shiny::icon("car"),
  #                                     shiny::icon("wrench")
  #                                   ),
  #                                   group = "grp"))
  # expect_is(m, "leaflet")
  # expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
})

test_that("popup-points-default", {
  ## POSIX* ##############
  m <- leaflet() %>% addTiles() %>%
                        addGlPoints(data = breweries91,
                                    popup = Sys.time(),
                                    group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))

  ## POSIX* - same length##############
  m <- leaflet() %>% addTiles() %>%
    addGlPoints(data = breweries91,
                popup = seq.POSIXt(Sys.time(), Sys.time()+1000,
                                   length.out = nrow(breweries91)),
                group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))

  ## Date ##############
  m <- leaflet() %>% addTiles() %>%
                        addGlPoints(data = breweries91,
                                    popup = Sys.Date(),
                                    group = "grp")
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
})

## LINES #################################
storms = suppressWarnings(st_cast(st_as_sf(atlStorms2005), "LINESTRING"))
storms = st_transform(storms, 4326)

## TODO - not working in RStudio Pane /In Zomm Mode it works / Shiny?
test_that("popup-lines-character", {
  ## Column Name Single ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPolylines(data = storms,
                   popup = "Name",
                   opacity = 1)
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
  expect_identical(read_json_str(m$x$calls[[2]]$args[[3]]), as.character(storms$Name))
  rm(m)

  m <- leaflet() %>% addTiles() %>%
    addGlPolylines(data = storms,
                   popup = storms$Name,
                   opacity = 1)
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
  expect_identical(read_json_str(m$x$calls[[2]]$args[[3]]), as.character(storms$Name))
  rm(m)


  ## Column Names ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPolylines(data = storms,
                   popup = c("Name", "MaxWind"),
                   opacity = 1)
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
  rm(m)

  ## Buttons / Icons / Emojis ?? ##############
  library(shiny)
  m <- leaflet() %>% addTiles() %>%
    addGlPolylines(data = storms,
                  popup = paste0(
                    storms$Name, ": ", storms$MaxWind,
                    actionButton("showmodal", "Expand to show more details",
                                 onclick = 'Shiny.onInputChange("button_click",  Math.random())')),
                  opacity = 1)
  expect_is(m, "leaflet")
  rm(m)

  m <- leaflet() %>% addTiles() %>%
    addGlPolylines(data = storms,
                  popup = shiny::icon("cog"),
                  opacity = 1)
  expect_is(m, "leaflet")
  rm(m)

  m <- leaflet() %>% addTiles() %>%
    addGlPolylines(data = storms,
                   src = TRUE,
                   popup = shiny::icon("cog"),
                   opacity = 1)
  expect_is(m, "leaflet")
  rm(m)

  # storms$icontext <- sample(c(htmltools::tagList(shiny::icon("cog")),
  #                             htmltools::tagList(shiny::icon("person")),
  #                             htmltools::tagList(shiny::icon("car")),
  #                             htmltools::tagList(shiny::icon("circle"))),
  #                             nrow(storms), replace = TRUE)
  # m <- leaflet() %>% addTiles() %>%
  #   addGlPolylines(data = storms,
  #                  popup = ~icontext,
  #                  opacity = 1)
  # expect_is(m, "leaflet")
  # rm(m)


  m <- leaflet() %>% addTiles() %>%
    addGlPolylines(data = storms,
                  popup = "\U0001f600",
                  opacity = 1)
  expect_is(m, "leaflet")
  rm(m)




  ## Single Random Character ##############
  m <- leaflet() %>% addTiles() %>%
                        addGlPolylines(data = storms,
                                       popup = "Text 1",
                                       opacity = 1)
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
  expect_identical(read_json_str(m$x$calls[[2]]$args[[3]]), rep("Text 1", nrow(storms)))
  rm(m)

  ## Multiple Random Characters - (wrong length) ##############
  m <- leaflet() %>% addTiles() %>%
                        addGlPolylines(data = storms,
                                    popup = c("Text 1", "Text 2"),
                                    opacity = 1)
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
  expect_identical(read_json_str(m$x$calls[[2]]$args[[3]]),
                   rep(c("Text 1","Text 2"), nrow(storms))[1:nrow(storms)])
  rm(m)

  ## Multiple Random Character (same length) ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPolylines(data = storms,
                   popup = rep("Text 1", nrow(storms)),
                   opacity = 1)
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
  expect_identical(read_json_str(m$x$calls[[2]]$args[[3]]), rep("Text 1", nrow(storms)))
})

test_that("popup-lines-table", {
  ## Data.frame ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPolylines(data = storms,
                   popup = as.data.frame(storms),
                   opacity = 1)
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
  m$x$calls[[2]]$args[[3]]

  m <- leaflet() %>% addTiles() %>%
    addGlPolylines(data = storms,
                   popup = as.data.frame(storms),
                   src = TRUE,
                   opacity = 1)
  expect_is(m, "leaflet")

  ## Data.frame - wrong length ##############
  m <- leaflet() %>% addTiles() %>%
                        addGlPolylines(data = storms,
                                       popup = as.data.frame(storms)[1:4,],
                                       opacity = 1)
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))

  m <- leaflet() %>% addTiles() %>%
    addGlPolylines(data = storms,
                   popup = as.data.frame(storms)[1:4,],
                   src = TRUE,
                   opacity = 1)
  expect_is(m, "leaflet")


  m <- leaflet() %>% addTiles() %>%
    addGlPolylines(data = storms,
                   popup = as.data.frame(storms)[1:4,],
                   weight = sample(0.5:5, nrow(storms), replace=TRUE),
                   src = TRUE,
                   opacity = 1)
  expect_is(m, "leaflet")


  ## Matrix ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPolylines(data = storms,
                   popup = as.matrix(as.data.frame(storms)),
                   opacity = 1)
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
})

test_that("popup-lines-spatial", {
  ## Simple Feature ##############
  library(sf)
  m <- leaflet() %>% addTiles() %>%
    addGlPolylines(data = storms,
                   popup = storms,
                   opacity = 1)

  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))

  skip_if_not_installed("sp")
  popups <- suppressWarnings(sf::as_Spatial(storms))
  ## SpatialLinesDataFrame ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPolylines(data = storms,
                   popup = popups,
                   opacity = 1)
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
})

test_that("popup-lines-formula", {
  ## Formula ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPolylines(data = storms,
                   popup = ~sprintf("<b>State</b>: %s<br>
                                 <b>Address</b>: %s<br>
                                 <b>Brauerei</b>: %s,",
                                    MinPress, MaxWind, Name),
                   opacity = 1)

  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
})

test_that("popup-lines-logical", {
  ## TRUE ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPolylines(data = storms,
                   popup = TRUE,
                   opacity = 1)
  expect_is(m, "leaflet")
  expect_true(m$x$calls[[2]]$args[[3]])

  ## FALSE #################
  m <- leaflet() %>% addTiles() %>%
    addGlPolylines(data = storms,
                   popup = FALSE,
                   opacity = 1)
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
  expect_true(m$x$calls[[2]]$args[[3]] == "null")

  ## NULL #################
  m <- leaflet() %>% addTiles() %>%
    addGlPolylines(data = storms,
                   popup = NULL,
                   opacity = 1)

  expect_is(m, "leaflet")
  expect_true(is.null(m$x$calls[[2]]$args[[3]]))
})

## POLYGONS #################################
gadm = suppressWarnings(st_cast(st_as_sf(gadmCHE), "POLYGON"))

test_that("popup-polygon-character", {
  ## Column Name Single ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPolygons(data = gadm,
                  popup = "HASC_1",
                  opacity = 1)
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
  expect_identical(read_json_str(m$x$calls[[2]]$args[[3]]), gadm$HASC_1)
  rm(m)

  m <- leaflet() %>% addTiles() %>%
    addGlPolygons(data = gadm,
                  popup = "HASC_1",
                  src = TRUE,
                  opacity = 1)
  expect_is(m, "leaflet")


  m <- leaflet() %>% addTiles() %>%
    addGlPolygons(data = gadm,
                  popup = gadm$HASC_1,
                  opacity = 1)
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
  expect_identical(read_json_str(m$x$calls[[2]]$args[[3]]), gadm$HASC_1)
  rm(m)

  m <- leaflet() %>% addTiles() %>%
    addGlPolygons(data = gadm,
                  popup = gadm$HASC_1,
                  src = TRUE,
                  opacity = 1)
  expect_is(m, "leaflet")


  ## Column Names ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPolygons(data = gadm,
                   popup = c("HASC_1", "NAME_0"),
                   opacity = 1)
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
  rm(m)

  m <- leaflet() %>% addTiles() %>%
    addGlPolygons(data = gadm,
                  popup = c("HASC_1", "NAME_0"),
                  src=TRUE,
                  opacity = 1)
  expect_is(m, "leaflet")

  ## Buttons / Icons / Emojis ##############
  library(shiny)
  m <- leaflet() %>% addTiles() %>%
    addGlPolygons(data = gadm,
                  popup = paste0(
                    gadm$NAME_1, ": ", gadm$NAME_0,
                    actionButton("showmodal", "Expand to show more details",
                                 onclick = 'Shiny.onInputChange("button_click",  Math.random())')),
                  opacity = 1)
  expect_is(m, "leaflet")
  rm(m)

  m <- leaflet() %>% addTiles() %>%
    addGlPolygons(data = gadm,
                  popup = shiny::icon("cog"),
                  opacity = 1)
  expect_is(m, "leaflet")
  rm(m)

  m <- leaflet() %>% addTiles() %>%
    addGlPolygons(data = gadm,
                  popup = shiny::icon("cog"),
                  src = TRUE,
                  opacity = 1)
  expect_is(m, "leaflet")
  rm(m)

  m <- leaflet() %>% addTiles() %>%
    addGlPolygons(data = gadm,
                  popup = "\U0001f600",
                  opacity = 1)
  expect_is(m, "leaflet")
  rm(m)




  ## Single Random Character ##############
  m <- leaflet() %>% addTiles() %>%
                        addGlPolygons(data = gadm,
                                       popup = "Text 1",
                                       opacity = 1)
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
  expect_identical(read_json_str(m$x$calls[[2]]$args[[3]]), rep("Text 1", nrow(gadm)))
  rm(m)

  ## Multiple Random Characters - (wrong length) ##############
  m <- leaflet() %>% addTiles() %>%
                        addGlPolygons(data = gadm,
                                       popup = c("Text 1", "Text 2"),
                                       opacity = 1)
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
  expect_identical(read_json_str(m$x$calls[[2]]$args[[3]]),
                   rep(c("Text 1","Text 2"), nrow(gadm))[1:nrow(gadm)])
  rm(m)

  ## Multiple Random Character (same length) ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPolygons(data = gadm,
                   popup = rep("Text 1", nrow(gadm)),
                   opacity = 1)
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
  expect_identical(read_json_str(m$x$calls[[2]]$args[[3]]), rep("Text 1", nrow(gadm)))
})

test_that("popup-polygon-table", {
  ## Data.frame ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPolygons(data = gadm,
                   popup = as.data.frame(gadm)[1:6],
                   opacity = 1)
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))

  ## Data.frame - wrong length ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPolygons(data = gadm,
                  popup = as.data.frame(gadm)[1, 1:4],
                  opacity = 1)
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))

  ## Matrix ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPolygons(data = gadm,
                   popup = as.matrix(as.data.frame(gadm)[1:6]),
                   opacity = 1)
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))

})

test_that("popup-polygon-spatial", {
  ## Simple Feature ##############
  library(sf)
  m <- leaflet() %>% addTiles() %>%
    addGlPolygons(data = gadm,
                  popup = gadm,
                  opacity = 1)

  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
  ## SpatialPolygonsDataFrame ##############
  skip_if_not_installed("sp")
  popups <- suppressWarnings(sf::as_Spatial(gadm))
  m <- leaflet() %>% addTiles() %>%
    addGlPolygons(data = gadm,
                   popup = popups,
                   opacity = 1)
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
})

test_that("popup-polygon-formula", {
  ## Formula ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPolygons(data = gadm,
                   popup = ~sprintf("<b>State</b>: %s<br>
                                 <b>Address</b>: %s<br>
                                 <b>Brauerei</b>: %s,",
                                    NAME_0, NAME_1, HASC_1),
                   opacity = 1)

  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
})

test_that("popup-polygon-logical", {
  ## TRUE ##############
  m <- leaflet() %>% addTiles() %>%
    addGlPolygons(data = gadm,
                   popup = TRUE,
                   opacity = 1)
  expect_is(m, "leaflet")
  expect_true(m$x$calls[[2]]$args[[3]])

  ## FALSE #################
  m <- leaflet() %>% addTiles() %>%
    addGlPolygons(data = gadm,
                   popup = FALSE,
                   opacity = 1)
  expect_is(m, "leaflet")
  expect_true(yyjsonr::validate_json_str(m$x$calls[[2]]$args[[3]]))
  expect_true(m$x$calls[[2]]$args[[3]] == "null")

  ## NULL #################
  m <- leaflet() %>% addTiles() %>%
    addGlPolygons(data = gadm,
                   popup = NULL,
                   opacity = 1)
  expect_is(m, "leaflet")
  expect_true(is.null(m$x$calls[[2]]$args[[3]]))
})

#' @title Add Data to a leaflet map using Leaflet.glify
#'
#' @description
#'   Leaflet.glify is a WebGL renderer plugin for leaflet. See
#'   \url{https://github.com/robertleeplummerjr/Leaflet.glify} for details
#'   and documentation.
#'
#' @inheritParams leaflet::addPolylines
#' @param data sf/sp point/polygon/line data to add to the map.
#' @param color Object representing the color. Can be of class integer, character with
#'   color names, HEX codes or random characters, factor, matrix, data.frame, list, json or formula.
#'   See the examples or \link{makeColorMatrix} for more information.
#' @param opacity feature opacity. Numeric between 0 and 1.
#'   Note: expect funny results if you set this to < 1.
#' @param radius point size in pixels.
#' @param popup Object representing the popup. Can be of type character with column names,
#'   formula, logical, data.frame or matrix, Spatial, list or JSON. If the length does not
#'   match the number of rows in the data, the popup vector is repeated to match the dimension.
#' @param weight line width/thickness in pixels for \code{addGlPolylines}.
#' @param src whether to pass data to the widget via file attachments.
#' @param pane A string which defines the pane of the layer. The default is \code{"overlayPane"}.
#' @param contextMenu a \code{\link[htmlwidgets]{JS}} function, that is passed to contextmenu.
#'  See the example in \code{./inst/examples/contextmenu.R}
#' @param ... Used to pass additional named arguments to \code{\link[yyjsonr]{write_json_str}} or
#'   \code{\link[yyjsonr]{write_geojson_str}} & to pass additional arguments to the
#'   underlying JavaScript functions. Typical use-cases include setting \code{'digits'} to
#'   round the point coordinates or to pass a different \code{'fragmentShaderSource'} to
#'   control the shape of the points. Use
#'   \itemize{
#'      \item{\code{'point'} (default) to render circles with a thin black outline}
#'      \item{\code{'simpleCircle'} for circles without outline}
#'      \item{\code{'square'} for squares without outline}
#'   }
#'   Additional arguments could be \code{'sensitivity'}, \code{'sensitivityHover'} or
#'   \code{'vertexShaderSource'}. See a full list at the
#'   \href{https://github.com/robertleeplummerjr/Leaflet.glify}{Leaflet.glify}
#'   repository.
#'
#'
#' @note
#'   MULTILINESTRINGs and MULTIPOLYGONs are currently not supported!
#'   Make sure you cast your data to LINESTRING or POLYGON first using:
#'   \itemize{
#'      \item{\code{sf::st_cast(data, "LINESTRING")}}
#'      \item{\code{sf::st_cast(data, "POLYGON")}}
#'   }
#'
#' @section Shiny Inputs:
#'   The objects created with \code{leafgl} send input values to Shiny as the
#'   user interacts with them. These events follow the pattern
#'   \code{input$MAPID_glify_EVENTNAME}.
#'   The following events are available:
#'
#'   \itemize{
#'     \item \strong{Click Events:}
#'       \code{input$MAPID_glify_click}
#'     \item \strong{Mouseover Events:}
#'       \code{input$MAPID_glify_mouseover}
#'     \item \strong{Mouseout Events:}
#'       \code{input$MAPID_glify_mouseout}
#'   }
#'
#'
#'   Each event returns a list containing:
#'   \itemize{
#'     \item \code{lat}: Latitude of the object or mouse cursor
#'     \item \code{lng}: Longitude of the object or mouse cursor
#'     \item \code{id}: The layerId, if any
#'     \item \code{group}: The group name of the object
#'     \item \code{data}: The properties of the feature
#'   }
#'
#' @describeIn addGlPoints Add Points to a leaflet map using Leaflet.glify
#' @examples \donttest{
#' library(leaflet)
#' library(leafgl)
#' library(sf)
#'
#' n = 1e5
#' df1 = data.frame(id = 1:n,
#'                  x = rnorm(n, 10, 1),
#'                  y = rnorm(n, 49, 0.8))
#' pts = st_as_sf(df1, coords = c("x", "y"), crs = 4326)
#' cols = topo.colors(nrow(pts))
#'
#' leaflet() %>%
#'   addProviderTiles(provider = providers$CartoDB.DarkMatter) %>%
#'   addGlPoints(data = pts, fillColor = cols, popup = TRUE)
#' }
#' @export addGlPoints
addGlPoints = function(map,
                       data,
                       fillColor = "#0033ff",
                       fillOpacity = 0.8,
                       radius = 10,
                       group = "glpoints",
                       popup = NULL,
                       label = NULL,
                       layerId = NULL,
                       src = FALSE,
                       pane = "overlayPane",
                       popupOptions = NULL,
                       labelOptions = NULL,
                       contextMenu = NULL,
                       ...) {

  # check data ##########
  if (missing(labelOptions)) labelOptions <- leaflet::labelOptions()
  if (missing(popupOptions)) popupOptions <- leaflet::popupOptions()

  if (is.null(group)) group = deparse(substitute(data))
  if (inherits(data, "Spatial")) data <- sf::st_as_sf(data)
  stopifnot(inherits(sf::st_geometry(data), c("sfc_POINT", "sfc_MULTIPOINT")))

  if (!is.null(layerId) && inherits(layerId, "formula"))
    layerId <- evalFormula(layerId, data)

  ## currently leaflet.glify only supports single (fill)opacity!
  fillOpacity = fillOpacity[1]

  # call SRC function ##############
  if (isTRUE(src)) {
    m = addGlPointsSrc(
      map = map
      , data = data
      , fillColor = fillColor
      , fillOpacity = fillOpacity
      , radius = radius
      , group = group
      , popup = popup
      , label = label
      , layerId = layerId
      , pane = pane
      , popupOptions = popupOptions
      , labelOptions = labelOptions
      , ...
    )
    return(m)
  }

  # get Bounds and ... #################
  dotopts = list(...)
  bounds = as.numeric(sf::st_bbox(data))

  # color ###########
  if (inherits(fillColor[1], "character") && startsWith(fillColor[1], "#")) {
    fillColor <- if(length(fillColor) == 1) {list(fillColor)} else { fillColor }
  } else {
    palette = "viridis"
    if ("palette" %in% names(dotopts)) {
      palette <- dotopts$palette
      dotopts$palette = NULL
    }
    fillColor <- makeColorMatrix(fillColor, data, palette = palette)
    if (ncol(fillColor) != 3) stop("only 3 column fillColor matrix supported so far")
    fillColor = as.data.frame(fillColor, stringsAsFactors = FALSE)
    colnames(fillColor) = c("r", "g", "b")
    fillColor = yyson_json_str(fillColor, digits = 3)
  }

  # label / popup ###########
  labels <- leaflet::evalFormula(label, data)
  if (!is.null(popup)) {
    htmldeps <- htmltools::htmlDependencies(popup)
    if (length(htmldeps) != 0) {
      map$dependencies = c(
        map$dependencies,
        htmldeps
      )
    }
    popup = yyson_json_str(makePopup(popup, data))
  } else {
    popup = NULL
  }

  # data ###########
  crds = sf::st_coordinates(data)[, c(2, 1)]
  if (length(dotopts) == 0) {
    jsonify_args = NULL
  } else {
    jsonify_args = try(
      match.arg(
        names(dotopts)
        , names(as.list(args(yyjsonr::opts_write_json)))
        , several.ok = TRUE
      )
      , silent = TRUE
    )
  }
  if (inherits(jsonify_args, "try-error")) jsonify_args = NULL
  data = do.call(yyson_json_str, c(list(crds), dotopts[jsonify_args]))
  class(data) <- "json"

  # dependencies
  map$dependencies = c(map$dependencies, glifyDependencies())

  # invoke leaflet method and zoom to bounds ###########
  map = leaflet::invokeMethod(
    map
    , leaflet::getMapData(map)
    , 'addGlifyPoints'
    , data
    , fillColor
    , popup
    , labels
    , fillOpacity
    , radius
    , group
    , layerId
    , dotopts
    , pane
    , popupOptions
    , labelOptions
    , contextMenu
  )

  leaflet::expandLimits(
    map,
    c(bounds[2], bounds[4]),
    c(bounds[1], bounds[3])
  )
}


### via src
addGlPointsSrc = function(map,
                          data,
                          fillColor = "#0033ff",
                          fillOpacity = 1,
                          radius = 10,
                          group = "glpoints",
                          popup = NULL,
                          label = NULL,
                          layerId = NULL,
                          pane = "overlayPane",
                          popupOptions = NULL,
                          labelOptions = NULL,
                          ...) {

  # get Bounds and ... #################
  dotopts = list(...)
  bounds = as.numeric(sf::st_bbox(data))

  # temp directories ############
  dir_data = tempfile(pattern = "glify_points_dat")
  dir.create(dir_data)
  dir_color = tempfile(pattern = "glify_points_col")
  dir.create(dir_color)
  dir_popup = tempfile(pattern = "glify_points_pop")
  dir.create(dir_popup)
  dir_radius = tempfile(pattern = "glify_points_rad")
  dir.create(dir_radius)
  dir_labels = tempfile(pattern = "glify_polylines_labl")
  dir.create(dir_labels)

  # data ############
  # data = sf::st_transform(data, 4326)
  crds = sf::st_coordinates(data)[, c(2, 1)]

  fl_data = paste0(dir_data, "/", group, "_data.js")
  pre = paste0('var data = data || {}; data["', group, '"] = ')
  writeLines(pre, fl_data)
  jsonify_args = try(
    match.arg(
      names(dotopts)
      , names(as.list(args(yyjsonr::opts_write_json)))
      , several.ok = TRUE
    )
    , silent = TRUE
  )
  if (inherits(jsonify_args, "try-error")) jsonify_args = NULL
  if (identical(jsonify_args, "x")) jsonify_args = NULL
  cat('[', do.call(yyson_json_str, c(list(crds), dotopts[jsonify_args])), '];',
      file = fl_data, sep = "", append = TRUE)

  map$dependencies = c(
    map$dependencies,
    glifyDependencies(TRUE),
    glifyDataAttachmentSrc(fl_data, group)
  )

  # color ############
  palette = "viridis"
  if ("palette" %in% names(dotopts)) {
    palette <- dotopts$palette
    dotopts$palette = NULL
  }
  fillColor <- makeColorMatrix(fillColor, data, palette = palette)
  if (ncol(fillColor) != 3) stop("only 3 column fillColor matrix supported so far")
  fillColor = as.data.frame(fillColor, stringsAsFactors = FALSE)
  colnames(fillColor) = c("r", "g", "b")

  if (nrow(fillColor) > 1) {
    fl_color = paste0(dir_color, "/", group, "_color.js")
    pre = paste0('var col = col || {}; col["', group, '"] = ')
    writeLines(pre, fl_color)
    cat('[', yyson_json_str(fillColor, digits = 3), '];',
        file = fl_color, append = TRUE)

    map$dependencies = c(map$dependencies, glifyAttachmentSrc(fl_color, group, "col"))
    fillColor = NULL
  }

  # labels ############
  if (!is.null(label)) {
    fl_label = paste0(dir_labels, "/", group, "_label.js")
    pre = paste0('var labs = labs || {}; labs["', group, '"] = ')
    writeLines(pre, fl_label)
    cat('[', yyson_json_str(leaflet::evalFormula(label, data)), '];',
        file = fl_label, append = TRUE)

    map$dependencies = c(map$dependencies, glifyAttachmentSrc(fl_label, group, "lab"))
    label = NULL
  }

  # popup ############
  if (!is.null(popup)) {
    htmldeps <- htmltools::htmlDependencies(popup)
    if (length(htmldeps) != 0) {
      map$dependencies = c(
        map$dependencies,
        htmldeps
      )
    }
    fl_popup = paste0(dir_popup, "/", group, "_popup.js")
    pre = paste0('var pops = pops || {}; pops["', group, '"] = ')
    writeLines(pre, fl_popup)
    cat('[', yyson_json_str(makePopup(popup, data)), '];',
        file = fl_popup, append = TRUE)

    map$dependencies = c(map$dependencies, glifyAttachmentSrc(fl_popup, group, "pop"))
    popup = NULL
  }

  # radius ############
  if (length(unique(radius)) > 1) {
    fl_radius = paste0(dir_radius, "/", layerId, "_radius.js")
    pre = paste0('var rad = rad || {}; rad["', layerId, '"] = ')
    writeLines(pre, fl_radius)
    cat('[', yyson_json_str(radius), '];',
        file = fl_radius, append = TRUE)

    map$dependencies = c(map$dependencies, glifyAttachmentSrc(fl_radius, group, "rad"))
    radius = NULL
  }

  # invoke method ###########
  map = leaflet::invokeMethod(
    map
    , leaflet::getMapData(map)
    , 'addGlifyPointsSrc'
    , fillColor
    , fillOpacity
    , radius
    , group
    , layerId
    , dotopts
    , pane
    , popupOptions
    , labelOptions
  )

  leaflet::expandLimits(
    map,
    c(bounds[2], bounds[4]),
    c(bounds[1], bounds[3])
  )
}


# ### via src ##############
# addGlPointsSrc2 = function(map,
#                            data,
#                            color = cbind(0, 0.2, 1),
#                            opacity = 1,
#                            weight = 10,
#                            group = "glpoints",
#                            popup = NULL,
#                            layerId = NULL,
#                            ...) {
#
#   if (is.null(group)) group = deparse(substitute(data))
#   if (inherits(data, "Spatial")) data <- sf::st_as_sf(data)
#   stopifnot(inherits(sf::st_geometry(data), c("sfc_POINT", "sfc_MULTIPOINT")))
#
#   # temp directories
#   dir_data = tempfile(pattern = "glify_points_dat")
#   dir.create(dir_data)
#   dir_color = tempfile(pattern = "glify_points_col")
#   dir.create(dir_color)
#   dir_popup = tempfile(pattern = "glify_points_pop")
#   dir.create(dir_popup)
#
#   # data
#   data = sf::st_transform(data, 4326)
#   crds = sf::st_coordinates(data)[, c(2, 1)]
#
#   grp1 = paste0(group, "_1")
#   grp2 = paste0(group, "_2")
#
#   fl_data1 = paste0(dir_data, "/", grp1, "_data.json")
#   fl_data2 = paste0(dir_data, "/", grp2, "_data.json")
#   pre1 = paste0('var data = data || {}; data["', grp1, '"] = ')
#   writeLines(pre1, fl_data1)
#   cat('[', yyson_json_str(crds[1:100, ], ...), '];',
#       file = fl_data1, sep = "", append = TRUE)
#   pre2 = paste0('var data = data || {}; data["', grp2, '"] = ')
#   writeLines(pre2, fl_data2)
#   cat('[', yyson_json_str(crds[101:nrow(crds), ], ...), '];',
#       file = fl_data2, sep = "", append = TRUE)
#
#   # color
#   if (ncol(color) != 3) stop("only 3 column color matrix supported so far")
#   color = as.data.frame(color, stringsAsFactors = FALSE)
#   colnames(color) = c("r", "g", "b")
#
#   fl_color = paste0(dir_color, "/", group, "_color.json")
#   pre = paste0('var col = col || {}; col["', group, '"] = ')
#   writeLines(pre, fl_color)
#   cat('[', yyson_json_str(color), '];',
#       file = fl_color, append = TRUE)
#
#   # popup
#   if (!is.null(popup)) {
#     fl_popup = paste0(dir_popup, "/", group, "_popup.json")
#     pre = paste0('var popup = popup || {}; popup["', group, '"] = ')
#     writeLines(pre, fl_popup)
#     cat('[', yyson_json_str(data[[popup]]), '];',
#         file = fl_popup, append = TRUE)
#   } else {
#     popup = NULL
#   }
#
#   # dependencies
#   map$dependencies = c(
#     map$dependencies,
#     glifyDependenciesSrc(TRUE),
#     glifyDataAttachmentSrc(fl_data1, grp1),
#     glifyDataAttachmentSrc(fl_data2, grp1, TRUE),
#     glifyAttachmentSrc(fl_color, group, "col")
#   )
#
#   if (!is.null(popup)) {
#     map$dependencies = c(map$dependencies, glifyAttachmentSrc(fl_color, group, "col"))
#   }
#
#   leaflet::invokeMethod(map, leaflet::getMapData(map), 'addGlifyPointsSrc2',
#                         group, opacity, weight, layerId)
#
# }
#
#
#
# ### via attachments
# addGlPointsFl = function(map,
#                          data,
#                          color = cbind(0, 0.2, 1),
#                          opacity = 1,
#                          weight = 10,
#                          group = "glpoints",
#                          popup = NULL,
#                          layerId = NULL,
#                          ...) {
#
#   if (is.null(group)) group = deparse(substitute(data))
#   if (inherits(data, "Spatial")) data <- sf::st_as_sf(data)
#   stopifnot(inherits(sf::st_geometry(data), c("sfc_POINT", "sfc_MULTIPOINT")))
#
#   # temp directories
#   dir_data = tempfile(pattern = "glify_points_dt")
#   dir.create(dir_data)
#   dir_color = tempfile(pattern = "glify_points_cl")
#   dir.create(dir_color)
#   dir_popup = tempfile(pattern = "glify_points_pop")
#   dir.create(dir_popup)
#
#   # data
#   data = sf::st_transform(data, 4326)
#   crds = sf::st_coordinates(data)[, c(2, 1)]
#
#   fl_data = paste0(dir_data, "/", group, "_data.json")
#   cat(yyson_json_str(crds, digits = 7), file = fl_data, append = FALSE)
#   data_var = paste0(group, "dt")
#
#   # color
#   if (ncol(color) != 3) stop("only 3 column color matrix supported so far")
#   color = as.data.frame(color, stringsAsFactors = FALSE)
#   colnames(color) = c("r", "g", "b")
#
#   jsn = yyson_json_str(color)
#   fl_color = paste0(dir_color, "/", group, "_color.json")
#   color_var = paste0(group, "cl")
#   cat(jsn, file = fl_color, append = FALSE)
#
#   # popup
#   if (!is.null(popup)) {
#     pop = yyson_json_str(data[[popup]])
#     fl_popup = paste0(dir_popup, "/", group, "_popup.json")
#     popup_var = paste0(group, "pop")
#     cat(pop, file = fl_popup, append = FALSE)
#   } else {
#     popup_var = NULL
#   }
#
#   # dependencies
#   map$dependencies = c(
#     map$dependencies,
#     glifyDependenciesFl(),
#     glifyDataAttachment(fl_data, group),
#     glifyColorAttachment(fl_color, group)
#   )
#
#   if (!is.null(popup)) {
#     map$dependencies = c(
#       map$dependencies,
#       glifyPopupAttachment(fl_popup, group)
#     )
#   }
#
#   leaflet::invokeMethod(map, leaflet::getMapData(map), 'addGlifyPointsFl',
#                         data_var, color_var, popup_var, opacity, weight, layerId)
#
# }

context("test-leafgl-addGlPoints")

test_that("addGlPoints works", {
  library(leaflet)
  library(sf)

  n = 1e3
  df1 = data.frame(id = 1:n,
                   id2 = n:1,
                   x = rnorm(n, 10, 1),
                   y = rnorm(n, 49, 0.8))
  pts = st_as_sf(df1, coords = c("x", "y"), crs = 4326)


  m = leaflet() %>%
    addGlPoints(data = pts, group = "pts", digits = 5,
                randomarg=TRUE,
                by ="row", numeric_dates=F)
  expect_is(m, "leaflet")


  m = leaflet() %>%
    addGlPoints(data = pts, group = "pts", digits = 5)
  expect_is(m, "leaflet")

  m = leaflet() %>%
    addGlPoints(data = pts, layerId = "someid", src = TRUE)
  expect_is(m, "leaflet")
  expect_identical(m$dependencies[[length(m$dependencies)-1]]$name, "glpointsdat")
  expect_identical(m$dependencies[[length(m$dependencies)]]$name, "glpointscol")

  m = leaflet() %>%
    addGlPoints(data = pts, group = NULL)
  expect_is(m, "leaflet")

  m = leaflet() %>%
    addGlPoints(data = pts, layerId = NULL, group = NULL, src = TRUE)
  expect_is(m, "leaflet")
  expect_identical(m$dependencies[[length(m$dependencies)-1]]$name, "ptsdat")
  expect_identical(m$dependencies[[length(m$dependencies)]]$name, "ptscol")

  m = leaflet() %>%
    addGlPoints(data = breweries91, src = TRUE)
  expect_is(m, "leaflet")
  expect_identical(m$dependencies[[length(m$dependencies)-1]]$name, "glpointsdat")
  expect_identical(m$dependencies[[length(m$dependencies)]]$name, "glpointscol")

  m = leaflet() %>%
    addGlPoints(data = breweries91, src = TRUE, radius = 5)
  expect_is(m, "leaflet")

  m = leaflet() %>%
    addGlPoints(data = breweries91, src = TRUE, radius = runif(nrow(breweries91), 1, 10))
  expect_is(m, "leaflet")
})

#' Use leafgl in shiny
#'
#' @param outputId output variable to read from
#' @param width,height the width and height of the map
#'
#' @details See \link[leaflet:map-shiny]{leaflet::leafletOutput} for details.
#'   \code{renderLeafgl} is only exported for consistency. You can just as well
#'   use \link[leaflet:map-shiny]{leaflet::renderLeaflet} (see example).
#'   \code{leafglOutput} on the other hand is needed as it will attach all
#'   necessary dependencies.
#'
#' @return A UI for rendering leafgl
#'
#' @rdname glify-shiny
#' @export
#'
#' @examples
#' if (interactive()) {
#' library(shiny)
#' library(leaflet)
#' library(leafgl)
#' library(sf)
#'
#' n = 1e4
#' df1 = data.frame(id = 1:n,
#'     x = rnorm(n, 10, 3),
#'     y = rnorm(n, 49, 1.8))
#' pts = st_as_sf(df1, coords = c("x", "y"), crs = 4326)
#'
#' m = leaflet() %>%
#'  addProviderTiles(provider = providers$CartoDB.DarkMatter) %>%
#'  addGlPoints(data = pts, group = "pts") %>%
#'  setView(lng = 10.5, lat = 49.5, zoom = 6) %>%
#'  addLayersControl(overlayGroups = "pts")
#'
#' ui <- fluidPage(
#'     leafglOutput("mymap")
#' )
#'
#' server <- function(input, output, session) {
#'     output$mymap <- renderLeaflet(m)
#' }
#'
#' shinyApp(ui, server)
#' }
#'

leafglOutput <- function(outputId, width = "100%", height = 400){
  tagList(
    leaflet::leafletOutput(outputId = outputId, width = width, height = height)
    , tags$script(glifyDependencies())
  )
}

# Just for consistency
#
#' @param expr An expression that generates an HTML widget
#' @param env The environment in which to evaluate expr.
#' @param quoted Is expr a quoted expression (with quote())?
#'   This is useful if you want to save an expression in a variable.
#'
#' @return A server function for rendering leafgl
#'
#' @rdname glify-shiny
#' @export
#'
renderLeafgl <- function(expr, env = parent.frame(), quoted = TRUE){
  leaflet::renderLeaflet(expr = expr, env = env, quoted = quoted)
}

context("test-leafgl-shiny")

test_that("shiny works", {
  ui <- leafglOutput("mymap")
  expect_is(ui, "shiny.tag.list")
  expect_is(ui, "list")
  expect_length(ui, 2)
  srv <- renderLeafgl({""})
  expect_is(srv, "shiny.render.function")
  expect_is(srv, "function")
})

#' @examples
#' library(leaflet)
#' library(leafgl)
#' library(sf)
#'
#' gadm = st_as_sf(gadmCHE)
#' gadm = st_cast(gadm, "POLYGON")
#' cols = grey.colors(nrow(gadm))
#'
#' leaflet() %>%
#'   addProviderTiles(provider = providers$CartoDB.DarkMatter) %>%
#'   addGlPolygons(data = gadm, color = cols, popup = TRUE)
#'
#' @describeIn addGlPoints Add Polygons to a leaflet map using Leaflet.glify
#' @aliases addGlPolygons
#' @export addGlPolygons
addGlPolygons = function(map,
                         data,
                         color = cbind(0, 0.2, 1),
                         fillColor = color,
                         fillOpacity = 0.8,
                         group = "glpolygons",
                         popup = NULL,
                         label = NULL,
                         layerId = NULL,
                         src = FALSE,
                         pane = "overlayPane",
                         stroke = TRUE,
                         popupOptions = NULL,
                         labelOptions = NULL,
                         contextMenu = NULL,
                         ...) {

  # check data ##########
  if (missing(labelOptions)) labelOptions <- leaflet::labelOptions()
  if (missing(popupOptions)) popupOptions <- leaflet::popupOptions()

  if (is.null(group)) group = deparse(substitute(data))
  if (inherits(data, "Spatial")) data <- sf::st_as_sf(data)
  stopifnot(inherits(sf::st_geometry(data), c("sfc_POLYGON", "sfc_MULTIPOLYGON")))
  if (inherits(sf::st_geometry(data), "sfc_MULTIPOLYGON"))
    stop("Can only handle POLYGONs, please cast your MULTIPOLYGON to POLYGON using sf::st_cast",
         call. = FALSE)

  if (!is.null(layerId) && inherits(layerId, "formula"))
    layerId <- evalFormula(layerId, data)

  ## currently leaflet.glify only supports single (fill)opacity!
  fillOpacity = fillOpacity[1]

  # call SRC function ##############
  if (isTRUE(src)) {
    m = addGlPolygonsSrc(
      map = map
      , data = data
      , color = color
      , fillColor = fillColor
      , fillOpacity = fillOpacity
      , group = group
      , popup = popup
      , label = label
      , layerId = layerId
      , pane = pane
      , stroke = stroke
      , popupOptions = popupOptions
      , labelOptions = labelOptions
      , ...
    )
    return(m)
  }

  # get Bounds and ... #################
  dotopts = list(...)
  bounds = as.numeric(sf::st_bbox(data))

  # fillColor ###########
  if (inherits(fillColor[1], "character") && startsWith(fillColor[1], "#")) {
    cols <- color
    cols <- if(length(cols) == 1) {list(cols)} else { cols }
  } else {
    palette = "viridis"
    if ("palette" %in% names(dotopts)) {
      palette <- dotopts$palette
      dotopts$palette = NULL
    }
    fillColor <- makeColorMatrix(fillColor, data, palette = palette)
    if (ncol(fillColor) != 3) stop("only 3 column fillColor matrix supported so far")
    fillColor = as.data.frame(fillColor, stringsAsFactors = FALSE)
    colnames(fillColor) = c("r", "g", "b")
    cols = yyson_json_str(fillColor, digits = 3)
  }

  # label / popup ###########
  labels <- leaflet::evalFormula(label, data)
  if (is.null(popup)) {
    # geom = sf::st_transform(sf::st_geometry(data), crs = 4326)
    geom = sf::st_geometry(data)
    data = sf::st_sf(id = 1:length(geom), geometry = geom)
  } else if (isTRUE(popup)) {
    ## Don't do anything. Pass all columns to JS
  } else {
    htmldeps <- htmltools::htmlDependencies(popup)
    if (length(htmldeps) != 0) {
      map$dependencies = c(
        map$dependencies,
        htmldeps
      )
    }
    popup = yyson_json_str(makePopup(popup, data))
    geom = sf::st_geometry(data)
    data = sf::st_sf(id = 1:length(geom), geometry = geom)
  }

  # data ###########
  if (length(dotopts) == 0) {
    geojsonsf_args = NULL
  } else {
    geojsonsf_args = try(
      match.arg(
        names(dotopts)
        , names(as.list(args(yyjsonr::opts_write_json)))
        , several.ok = TRUE
      )
      , silent = TRUE
    )
    if (inherits(geojsonsf_args, "try-error")) geojsonsf_args = NULL
    if (identical(geojsonsf_args, "sf")) geojsonsf_args = NULL
  }
  data = do.call(yyson_geojson_str, c(list(data), "json_opts" = list(dotopts[geojsonsf_args])))

  # dependencies
  map$dependencies = c(map$dependencies, glifyDependencies())

  # invoke leaflet method and zoom to bounds ###########
  map = leaflet::invokeMethod(
    map
    , leaflet::getMapData(map)
    , 'addGlifyPolygons'
    , data
    , cols
    , popup
    , labels
    , fillOpacity
    , group
    , layerId
    , dotopts
    , pane
    , stroke
    , popupOptions
    , labelOptions
    , contextMenu
  )

  leaflet::expandLimits(
    map,
    c(bounds[2], bounds[4]),
    c(bounds[1], bounds[3])
  )
}



### via src
addGlPolygonsSrc = function(map,
                            data,
                            color = cbind(0, 0.2, 1),
                            fillColor = color,
                            fillOpacity = 0.6,
                            group = "glpolygons",
                            popup = NULL,
                            label = NULL,
                            layerId = NULL,
                            pane = "overlayPane",
                            stroke = TRUE,
                            popupOptions = NULL,
                            labelOptions = NULL,
                            ...) {

  # get Bounds and ... #################
  dotopts = list(...)
  bounds = as.numeric(sf::st_bbox(data))

  # temp directories ############
  dir_data = tempfile(pattern = "glify_polygons_dat")
  dir.create(dir_data)
  dir_color = tempfile(pattern = "glify_polygons_col")
  dir.create(dir_color)
  dir_popup = tempfile(pattern = "glify_polygons_pop")
  dir.create(dir_popup)
  dir_labels = tempfile(pattern = "glify_polylines_labl")
  dir.create(dir_labels)

  # data ############
  data_orig <- data
  geom = sf::st_geometry(data)
  data = sf::st_sf(id = 1:length(geom), geometry = geom)

  fl_data = paste0(dir_data, "/", group, "_data.js")
  pre = paste0('var data = data || {}; data["', group, '"] = ')
  writeLines(pre, fl_data)
  if (length(dotopts) != 0) {
    jsonify_args = try(
      match.arg(
        names(dotopts)
        , names(as.list(args(yyjsonr::opts_write_json)))
        , several.ok = TRUE
      )
      , silent = TRUE
    )
    if (inherits(jsonify_args, "try-error")) jsonify_args = NULL
    if (identical(jsonify_args, "sf")) jsonify_args = NULL
    dotoptslist <- c("json_opts" = list(dotopts[jsonify_args]))
  } else {
    dotoptslist <- NULL
  }
  cat('[', do.call(yyson_geojson_str, c(list(data), dotoptslist)), '];',
      file = fl_data, sep = "", append = TRUE)

  map$dependencies = c(
    map$dependencies,
    glifyDependencies(TRUE),
    glifyDataAttachmentSrc(fl_data, group)
  )

  # color ############
  palette = "viridis"
  if ("palette" %in% names(dotopts)) {
    palette <- dotopts$palette
    dotopts$palette = NULL
  }
  fillColor <- makeColorMatrix(fillColor, data_orig, palette = palette)
  if (ncol(fillColor) != 3) stop("only 3 column fillColor matrix supported so far")
  fillColor = as.data.frame(fillColor, stringsAsFactors = FALSE)
  colnames(fillColor) = c("r", "g", "b")

  if (nrow(fillColor) > 1) {
    fl_color = paste0(dir_color, "/", group, "_color.js")
    pre = paste0('var col = col || {}; col["', group, '"] = ')
    writeLines(pre, fl_color)
    cat('[', yyson_json_str(fillColor, digits = 3), '];',
        file = fl_color, append = TRUE)

    map$dependencies = c(map$dependencies, glifyAttachmentSrc(fl_color, group, "col"))

    fillColor = NULL
  }

  # labels ############
  if (!is.null(label)) {
    fl_label = paste0(dir_labels, "/", group, "_label.js")
    pre = paste0('var labs = labs || {}; labs["', group, '"] = ')
    writeLines(pre, fl_label)
    cat('[', yyson_json_str(leaflet::evalFormula(label, data_orig)), '];',
        file = fl_label, append = TRUE)

    map$dependencies = c(map$dependencies, glifyAttachmentSrc(fl_label, group, "lab"))
    label = NULL
  }

  # popup ############
  if (!is.null(popup)) {
    htmldeps <- htmltools::htmlDependencies(popup)
    if (length(htmldeps) != 0) {
      map$dependencies = c(
        map$dependencies,
        htmldeps
      )
    }
    fl_popup = paste0(dir_popup, "/", group, "_popup.js")
    pre = paste0('var pops = pops || {}; pops["', group, '"] = ')
    writeLines(pre, fl_popup)
    cat('[', yyson_json_str(makePopup(popup, data_orig)), '];',
        file = fl_popup, append = TRUE)

    map$dependencies = c(map$dependencies, glifyAttachmentSrc(fl_popup, group, "pop"))
    popup = NULL
  }

  # invoke method ###########
  map = leaflet::invokeMethod(
    map
    , leaflet::getMapData(map)
    , 'addGlifyPolygonsSrc'
    , fillColor
    , fillOpacity
    , group
    , layerId
    , dotopts
    , pane
    , stroke
    , popupOptions
    , labelOptions
  )

  leaflet::expandLimits(
    map,
    c(bounds[2], bounds[4]),
    c(bounds[1], bounds[3])
  )

}


# ### via attachments ############
# addGlPolygonsFl = function(map,
#                            data,
#                            color = cbind(0, 0.2, 1),
#                            opacity = 0.6,
#                            weight = 10,
#                            group = "glpolygons",
#                            popup = NULL,
#                            ...) {
#
#   if (is.null(group)) group = deparse(substitute(data))
#   if (inherits(data, "Spatial")) data <- sf::st_as_sf(data)
#   stopifnot(inherits(sf::st_geometry(data), c("sfc_POLYGON", "sfc_MULTIPOLYGON")))
#   if (inherits(sf::st_geometry(data), "sfc_MULTIPOLYGON"))
#     stop("Can only handle POLYGONs, please cast your MULTIPOLYGON to POLYGON using sf::st_cast")
#
#   # temp directories
#   dir_data = tempfile(pattern = "glify_polygons_dt")
#   dir.create(dir_data)
#   dir_color = tempfile(pattern = "glify_polygons_cl")
#   dir.create(dir_color)
#   # dir_popup = tempfile(pattern = "glify_polygons_pop")
#   # dir.create(dir_popup)
#
#   # data
#   geom = sf::st_transform(sf::st_geometry(data), crs = 4326)
#   if (is.null(popup)) {
#     data = sf::st_sf(id = 1:length(geom),
#                      geometry = geom)
#   } else {
#     data = sf::st_transform(data[, popup], crs = 4326)
#   }
#   # crds = sf::st_coordinates(data)[, c(2, 1)]
#
#   fl_data = paste0(dir_data, "/", group, "_data.json")
#   cat(geojsonsf::sf_geojson(data), file = fl_data, append = FALSE)
#   data_var = paste0(group, "dt")
#
#   # color
#   if (ncol(color) != 3) stop("only 3 column color matrix supported so far")
#   color = as.data.frame(color, stringsAsFactors = FALSE)
#   colnames(color) = c("r", "g", "b")
#
#   jsn = jsonify::to_json(color)
#   fl_color = paste0(dir_color, "/", group, "_color.json")
#   color_var = paste0(group, "cl")
#   cat(jsn, file = fl_color, append = FALSE)
#
#   # popup
#   # if (!is.null(popup)) {
#   #   pop = jsonlite::toJSON(data[[popup]])
#   #   fl_popup = paste0(dir_popup, "/", group, "_popup.json")
#   #   popup_var = paste0(group, "pop")
#   #   cat(pop, file = fl_popup, append = FALSE)
#   # } else {
#   #   popup_var = NULL
#   # }
#
#   # dependencies
#   map$dependencies = c(
#     map$dependencies,
#     glifyDependencies(),
#     glifyDataAttachment(fl_data, group),
#     glifyColorAttachment(fl_color, group)
#   )
#
#   # if (!is.null(popup)) {
#   #   map$dependencies = c(
#   #     map$dependencies,
#   #     glifyPopupAttachment(fl_popup, group)
#   #   )
#   # }
#
#   leaflet::invokeMethod(map, leaflet::getMapData(map), 'addGlifyPolygonsFl',
#                         data_var, color_var, popup, opacity)
#
# }

context("test-leafgl-addGlPolygons")

test_that("addGlPolygons works", {
  library(leaflet)
  library(leafgl)
  library(sf)

  gadm = st_as_sf(gadmCHE)
  single_poly <- suppressWarnings(st_cast(gadm, "POLYGON"))

  m = leaflet() %>%
    addGlPolygons(data = single_poly,
                  group = "pls", digits = 5)
  expect_is(m, "leaflet")

  # Group = NULL #######
  m = leaflet() %>%
    addGlPolygons(data = single_poly, group = NULL, digits = 5)
  expect_is(m, "leaflet")

  m = leaflet() %>%
    addGlPolygons(data = single_poly, group = NULL, src = TRUE, digits = 5)
  expect_is(m, "leaflet")

  ## Multi #########
  expect_error(
    leaflet() %>%
      addGlPolygons(data = gadm, digits = 5))
  expect_error(
    leaflet() %>%
      addGlPolygons(data = gadm, src = TRUE, digits = 5))

  ## Spatial Data #########
  skip_if_not_installed("sp")
  spatialdf <- as(single_poly, "Spatial")
  m = leaflet() %>%
    addGlPolygons(data = spatialdf, digits = 5)
  expect_is(m, "leaflet")

  m = leaflet() %>%
    addGlPolygons(data = spatialdf, src = TRUE, digits = 5)
  expect_is(m, "leaflet")
})

# dependencies
glifyDependencies = function(src = FALSE) {
  src <- ifelse(src, "Src", "")
  list(
    htmltools::htmlDependency(
      "Leaflet.glify",
      '3.3.1',
      system.file("htmlwidgets/Leaflet.glify", package = "leafgl"),
      script = c(
        "GlifyUtils.js"
        , paste0("addGlifyPoints", src, ".js")
        , paste0("addGlifyPolygons", src, ".js")
        , paste0("addGlifyPolylines", src, ".js")
        , "glify-browser.js"
      )
    )
  )
}

glifyDataAttachmentSrc = function(fl_data, group, async = FALSE) {
  data_dir <- dirname(fl_data)
  data_file <- basename(fl_data)
  if (async) {
    list(
      htmltools::htmlDependency(
        name = paste0("pts_2", "dat"),
        version = 1,
        src = "",
        head = paste0('<script src="lib/', paste0(group, 'dat-1/'), data_file, '" async></script>')
      )
    )
  } else {
    list(
      htmltools::htmlDependency(
        name = paste0(group, "dat"),
        version = 1,
        src = c(file = data_dir),
        script = list(data_file)
      )
    )
  }
}

glifyAttachmentSrc <- function(fl, group, type) {
  valid_types <- c("col", "pop", "lab", "rad")
  if (!type %in% valid_types) {
    stop("Invalid type. Valid types are: col, pop, lab, rad.")
  }
  data_dir <- dirname(fl)
  data_file <- basename(fl)
  list(
    htmltools::htmlDependency(
      name = paste0(group, type),
      version = 1,
      src = c(file = data_dir),
      script = list(data_file)
    )
  )
}


# helpers
yyson_json_str <- function(x, ...) {
  dt <- yyjsonr::write_json_str(x, ...)
  class(dt) <- "json"
  dt
}
yyson_geojson_str <- function(x, ...) {
  dt <- yyjsonr::write_geojson_str(x, ...)
  class(dt) <- "json"
  dt
}


## Not used ##########
# glifyDependenciesFl = function() {
#   list(
#     htmltools::htmlDependency(
#       "Leaflet.glify",
#       '2.2.0',
#       system.file("htmlwidgets/Leaflet.glify", package = "leafgl"),
#       script = c(
#         "GlifyUtils.js"
#         , "addGlifyPoints.js"
#         , "addGlifyPolygonsFl.js"
#         , "addGlifyPolylines.js"
#         , "glify-browser.js"
#       )
#     )
#   )
# }
# glifyDataAttachment = function(fl_data, group) {
#   data_dir <- dirname(fl_data)
#   data_file <- basename(fl_data)
#   list(
#     htmltools::htmlDependency(
#       name = paste0(group, "dt"),
#       version = 1,
#       src = c(file = data_dir),
#       attachment = list(data_file)
#     )
#   )
# }
# glifyColorAttachment = function(fl_color, group) {
#   data_dir <- dirname(fl_color)
#   data_file <- basename(fl_color)
#   list(
#     htmltools::htmlDependency(
#       name = paste0(group, "cl"),
#       version = 1,
#       src = c(file = data_dir),
#       attachment = list(data_file)
#     )
#   )
# }
# glifyPopupAttachment = function(fl_popup, group) {
#   data_dir <- dirname(fl_popup)
#   data_file <- basename(fl_popup)
#   list(
#     htmltools::htmlDependency(
#       name = paste0(group, "pop"),
#       version = 1,
#       src = c(file = data_dir),
#       attachment = list(data_file)
#     )
#   )
# }

#' @examples
#' library(leaflet)
#' library(leafgl)
#' library(sf)
#'
#' storms = st_as_sf(atlStorms2005)
#' cols = heat.colors(nrow(storms))
#'
#' leaflet() %>%
#'   addProviderTiles(provider = providers$CartoDB.Positron) %>%
#'   addGlPolylines(data = storms, color = cols, popup = TRUE, opacity = 1)
#'
#' @describeIn addGlPoints Add Lines to a leaflet map using Leaflet.glify
#' @aliases addGlPolylines
#' @export addGlPolylines
addGlPolylines = function(map,
                          data,
                          color = cbind(0, 0.2, 1),
                          opacity = 0.6,
                          group = "glpolylines",
                          popup = NULL,
                          label = NULL,
                          weight = 1,
                          layerId = NULL,
                          src = FALSE,
                          pane = "overlayPane",
                          popupOptions = NULL,
                          labelOptions = NULL,
                          contextMenu = NULL,
                          ...) {

  # check data ##########
  if (missing(labelOptions)) labelOptions <- leaflet::labelOptions()
  if (missing(popupOptions)) popupOptions <- leaflet::popupOptions()

  if (is.null(group)) group = deparse(substitute(data))
  if (inherits(data, "Spatial")) data <- sf::st_as_sf(data)
  stopifnot(inherits(sf::st_geometry(data), c("sfc_LINESTRING", "sfc_MULTILINESTRING")))
  if (inherits(sf::st_geometry(data), "sfc_MULTILINESTRING"))
    stop("Can only handle LINESTRINGs, please cast your MULTILINESTRING to LINESTRING using sf::st_cast",
         call. = FALSE)

  if (!is.null(layerId) && inherits(layerId, "formula"))
    layerId <- evalFormula(layerId, data)

  ## currently leaflet.glify only supports single (fill)opacity!
  opacity = opacity[1]

  # call SRC function ##############
  if (isTRUE(src)) {
    m = addGlPolylinesSrc(
      map = map
      , data = data
      , color = color
      , opacity = opacity
      , group = group
      , popup = popup
      , label = label
      , weight = weight
      , layerId = layerId
      , pane = pane
      , popupOptions = popupOptions
      , labelOptions = labelOptions
      , ...
    )
    return(m)
  }

  # get Bounds and ... #################
  dotopts = list(...)
  bounds = as.numeric(sf::st_bbox(data))

  # color ########
  if (inherits(color[1], "character") && startsWith(color[1], "#")) {
    cols <- color
    cols <- if(length(cols) == 1) {list(cols)} else { cols }
  } else {
    palette = "viridis"
    if ("palette" %in% names(dotopts)) {
      palette <- dotopts$palette
      dotopts$palette = NULL
    }
    color <- makeColorMatrix(color, data, palette = palette)
    if (ncol(color) != 3) stop("only 3 column color matrix supported so far")
    color = as.data.frame(color, stringsAsFactors = FALSE)
    colnames(color) = c("r", "g", "b")
    cols = yyson_json_str(color, digits = 3)
  }

  # label / popup ########
  labels <- leaflet::evalFormula(label, data)
  if (is.null(popup)) {
    geom = sf::st_geometry(data)
    data = sf::st_sf(id = 1:length(geom), geometry = geom)
  } else if (isTRUE(popup)) {
    ## Don't do anything. Pass all columns to JS
  } else {
    htmldeps <- htmltools::htmlDependencies(popup)
    if (length(htmldeps) != 0) {
      map$dependencies = c(
        map$dependencies,
        htmldeps
      )
    }
    popup = yyson_json_str(makePopup(popup, data))
    geom = sf::st_geometry(data)
    data = sf::st_sf(id = 1:length(geom), geometry = geom)
  }

  # data ########
  if (length(dotopts) == 0) {
    geojsonsf_args = NULL
  } else {
    geojsonsf_args = try(
      match.arg(
        names(dotopts)
        , names(as.list(args(yyjsonr::opts_write_json)))
        , several.ok = TRUE
      )
      , silent = TRUE
    )
    if (inherits(geojsonsf_args, "try-error")) geojsonsf_args = NULL
    if (identical(geojsonsf_args, "sf")) geojsonsf_args = NULL
  }
  data = do.call(yyson_geojson_str, c(list(data), "json_opts" = list(dotopts[geojsonsf_args])))

  # dependencies
  map$dependencies = c(map$dependencies, glifyDependencies())

  # invoke leaflet method and zoom to bounds ########
  map = leaflet::invokeMethod(
    map
    , leaflet::getMapData(map)
    , 'addGlifyPolylines'
    , data
    , cols
    , popup
    , labels
    , opacity
    , group
    , weight
    , layerId
    , dotopts
    , pane
    , popupOptions
    , labelOptions
    , contextMenu
  )

  leaflet::expandLimits(
    map,
    c(bounds[2], bounds[4]),
    c(bounds[1], bounds[3])
  )
}


### via src
addGlPolylinesSrc = function(map,
                             data,
                             color = cbind(0, 0.2, 1),
                             opacity = 0.8,
                             group = "glpolygons",
                             popup = NULL,
                             label = NULL,
                             weight = 1,
                             layerId = NULL,
                             pane = "overlayPane",
                             popupOptions = NULL,
                             labelOptions = NULL,
                             ...) {

  # get Bounds and ... #################
  dotopts = list(...)
  bounds = as.numeric(sf::st_bbox(data))

  # temp directories ############
  dir_data = tempfile(pattern = "glify_polylines_dat")
  dir.create(dir_data)
  dir_color = tempfile(pattern = "glify_polylines_col")
  dir.create(dir_color)
  dir_weight = tempfile(pattern = "glify_polylines_wgt")
  dir.create(dir_weight)
  dir_popup = tempfile(pattern = "glify_polylines_pop")
  dir.create(dir_popup)
  dir_labels = tempfile(pattern = "glify_polylines_labl")
  dir.create(dir_labels)

  # data ############
  data_orig <- data
  geom = sf::st_geometry(data)
  data = sf::st_sf(id = 1:length(geom), geometry = geom)

  fl_data = paste0(dir_data, "/", group, "_data.js")
  pre = paste0('var data = data || {}; data["', group, '"] = ')
  writeLines(pre, fl_data)
  if (length(dotopts) != 0) {
    jsonify_args = try(
      match.arg(
        names(dotopts)
        , names(as.list(args(yyjsonr::opts_write_json)))
        , several.ok = TRUE
      )
      , silent = TRUE
    )
    if (inherits(jsonify_args, "try-error")) jsonify_args = NULL
    if (identical(jsonify_args, "sf")) jsonify_args = NULL

    dotoptslist <- c("json_opts" = list(dotopts[jsonify_args]))
  } else {
    dotoptslist <- NULL
  }
  cat('[', do.call(yyson_geojson_str, c(list(data), dotoptslist)), '];',
      file = fl_data, sep = "", append = TRUE)

  map$dependencies = c(
    map$dependencies,
    glifyDependencies(TRUE),
    glifyDataAttachmentSrc(fl_data, group)
  )

  # color ############
  palette = "viridis"
  if ("palette" %in% names(dotopts)) {
    palette <- dotopts$palette
    dotopts$palette = NULL
  }
  color <- makeColorMatrix(color, data_orig, palette = palette)
  if (ncol(color) != 3) stop("only 3 column color matrix supported so far")
  color = as.data.frame(color, stringsAsFactors = FALSE)
  colnames(color) = c("r", "g", "b")
  if (nrow(color) > 1) {
    fl_color = paste0(dir_color, "/", group, "_color.js")
    pre = paste0('var col = col || {}; col["', group, '"] = ')
    writeLines(pre, fl_color)
    cat('[', yyson_json_str(color, digits = 3), '];',
        file = fl_color, append = TRUE)

    map$dependencies = c(map$dependencies, glifyAttachmentSrc(fl_color, group, "col"))
    color = NULL
  }

  # labels ############
  if (!is.null(label)) {
    fl_label = paste0(dir_labels, "/", group, "_label.js")
    pre = paste0('var labs = labs || {}; labs["', group, '"] = ')
    writeLines(pre, fl_label)
    cat('[', yyson_json_str(leaflet::evalFormula(label, data_orig)), '];',
        file = fl_label, append = TRUE)

    map$dependencies = c(map$dependencies, glifyAttachmentSrc(fl_label, group, "lab"))
    label = NULL
  }

  # popup ############
  if (!is.null(popup)) {
    htmldeps <- htmltools::htmlDependencies(popup)
    if (length(htmldeps) != 0) {
      map$dependencies = c(
        map$dependencies,
        htmldeps
      )
    }
    fl_popup = paste0(dir_popup, "/", group, "_popup.js")
    pre = paste0('var pops = pops || {}; pops["', group, '"] = ')
    writeLines(pre, fl_popup)
    cat('[', yyson_json_str(makePopup(popup, data_orig)), '];',
        file = fl_popup, append = TRUE)

    map$dependencies = c(map$dependencies, glifyAttachmentSrc(fl_popup, group, "pop"))
    popup = NULL
  }

  # weight ############
  if (length(unique(weight)) > 1) {
    fl_weight = paste0(dir_weight, "/", group, "_weight.js")
    pre = paste0('var wgt = wgt || {}; wgt["', group, '"] = ')
    writeLines(pre, fl_weight)
    cat('[', yyson_json_str(weight), '];',
        file = fl_weight, append = TRUE)

    map$dependencies = c(
      map$dependencies, glifyAttachmentSrc(fl_weight, group, "rad"))
    weight = NULL
  }

  # invoke method ###########
  map = leaflet::invokeMethod(
    map
    , leaflet::getMapData(map)
    , 'addGlifyPolylinesSrc'
    , color
    , opacity
    , group
    , weight
    , layerId
    , dotopts
    , pane
    , popupOptions
    , labelOptions
  )

  leaflet::expandLimits(
    map,
    c(bounds[2], bounds[4]),
    c(bounds[1], bounds[3])
  )
}

context("test-leafgl-addGlPolylines")

test_that("addGlPolylines works", {
  library(leaflet)
  library(leafgl)
  library(sf)

  storms = st_as_sf(atlStorms2005)

  m = leaflet() %>%
    addGlPolylines(data = suppressWarnings(st_cast(storms, "LINESTRING")),
                   group = "pls", digits = 5)
  expect_is(m, "leaflet")

  # Group = NULL #######
  m = leaflet() %>%
    addGlPolylines(data = storms, group = NULL, digits = 5)
  expect_is(m, "leaflet")

  m = leaflet() %>%
    addGlPolylines(data = storms, group = NULL, src = TRUE, digits = 5)
  expect_is(m, "leaflet")


  ## Spatial Data #########
  lines = atlStorms2005
  m = leaflet() %>%
    addGlPolylines(data = lines, digits = 5)
  expect_is(m, "leaflet")

  m = leaflet() %>%
    addGlPolylines(data = lines, src = TRUE, digits = 5)
  expect_is(m, "leaflet")


  ## Multi #########
  multi <- st_cast(storms, "MULTILINESTRING")
  expect_error(
    leaflet() %>%
      addGlPolylines(data = multi, digits = 5))
  expect_error(
    leaflet() %>%
      addGlPolylines(data = multi, src = TRUE, digits = 5))




  # m = leaflet() %>%
  #   addGlPolylines(data = storms, digits = 5,
  #                  popup = ~sprintf("Name: %s<br>%s", Name,
  #                                  shiny::actionButton("id", "Act")))
  # expect_is(m, "leaflet")
  # m$dependencies
  #
  # leaflet() %>%
  #   addPolylines(data = storms,
  #                  popup = ~sprintf("Name: %s<br>%s", Name,
  #                                   shiny::icon("cog")))



})

#' Remove Leaflet.Glify elements from a map
#'
#' Remove one or more features from a map, identified by `layerId`;
#' or, clear all features of the given group.
#'
#' @inheritParams leaflet::removeShape
#' @return the new `map` object
#'
#' @name remove
#' @export
removeGlPoints <- function(map, layerId) {
  leaflet::invokeMethod(map, NULL, "removeGlPoints", layerId)
}

#' @rdname remove
#' @export
removeGlPolylines <- function(map, layerId) {
  leaflet::invokeMethod(map, NULL, "removeGlPolylines", layerId)
}

#' @rdname remove
#' @export
removeGlPolygons <- function(map, layerId) {
  leaflet::invokeMethod(map, NULL, "removeGlPolygons", layerId)
}

#' @rdname remove
#' @export
clearGlLayers <- function(map) {
  leaflet::invokeMethod(map, NULL, "clearGlLayers")
}

#' @rdname remove
#' @export
clearGlGroup <- function(map, group) {
  leaflet::invokeMethod(map, NULL, "clearGlGroup", group)
}

context("test-leafgl-remove_clear")

library(leaflet)
library(sf)

n = 1e2
df1 = data.frame(id = 1:n, id2 = n:1,
                 x = rnorm(n, 10, 1),
                 y = rnorm(n, 49, 0.8))
pts = st_as_sf(df1, coords = c("x", "y"), crs = 4326)

lines = suppressWarnings(st_cast(st_as_sf(atlStorms2005), "LINESTRING"))
lines = st_transform(lines, 4326)[1:100,]

polys <- suppressWarnings(st_cast(st_as_sf(gadmCHE), "POLYGON"))

test_that("remove / clear", {
  ## Remove Points ########
  m <- leaflet() %>%
    addGlPoints(data = pts,
                layerId = "ptsid") %>%
    removeGlPoints("ptsid")

  expect_true(m$x$calls[[length(m$x$calls)]]$method == "removeGlPoints")
  expect_true(m$x$calls[[length(m$x$calls)]]$args[[1]] == "ptsid")

  ## Remove Lines ########
  m <- leaflet() %>%
    addGlPolylines(data = lines,
                   layerId = "lnsid") %>%
    removeGlPolylines("lnsid")

  expect_true(m$x$calls[[length(m$x$calls)]]$method == "removeGlPolylines")
  expect_true(m$x$calls[[length(m$x$calls)]]$args[[1]] == "lnsid")

  ## Remove Polygons ########
  m <- leaflet() %>%
    addGlPolygons(data = polys,
                  layerId = "polyid") %>%
    removeGlPolygons("polyid")

  expect_true(m$x$calls[[length(m$x$calls)]]$method == "removeGlPolygons")
  expect_true(m$x$calls[[length(m$x$calls)]]$args[[1]] == "polyid")

  ## Clear Layers ########
  m <- leaflet() %>%
    addGlPolygons(data = polys,
                  layerId = "polyid") %>%
    addGlPolylines(data = lines,
                   layerId = "lnsid") %>%
    addGlPoints(data = pts,
                layerId = "ptsid") %>%
    clearGlLayers()

  expect_true(m$x$calls[[length(m$x$calls)]]$method == "clearGlLayers")
  expect_identical(m$x$calls[[length(m$x$calls)]]$args, list())

})

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @importFrom leaflet leafletOutput renderLeaflet evalFormula
#' @importFrom htmltools tagList tags htmlDependencies
## usethis namespace: end
NULL

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-popup.R
\name{checkDimPop}
\alias{checkDimPop}
\title{checkDim}
\usage{
checkDimPop(x, data)
}
\arguments{
\item{x}{The popup vector}

\item{data}{The dataset}
}
\description{
Check the length of the popup vector. It must match the
  number of rows of the dataset.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/glify-lines.R, R/glify-points.R,
%   R/glify-polygons.R
\name{addGlPolylines}
\alias{addGlPolylines}
\alias{addGlPoints}
\alias{addGlPolygons}
\title{Add Data to a leaflet map using Leaflet.glify}
\usage{
addGlPolylines(
  map,
  data,
  color = cbind(0, 0.2, 1),
  opacity = 0.6,
  group = "glpolylines",
  popup = NULL,
  label = NULL,
  weight = 1,
  layerId = NULL,
  src = FALSE,
  pane = "overlayPane",
  popupOptions = NULL,
  labelOptions = NULL,
  contextMenu = NULL,
  ...
)

addGlPoints(
  map,
  data,
  fillColor = "#0033ff",
  fillOpacity = 0.8,
  radius = 10,
  group = "glpoints",
  popup = NULL,
  label = NULL,
  layerId = NULL,
  src = FALSE,
  pane = "overlayPane",
  popupOptions = NULL,
  labelOptions = NULL,
  contextMenu = NULL,
  ...
)

addGlPolygons(
  map,
  data,
  color = cbind(0, 0.2, 1),
  fillColor = color,
  fillOpacity = 0.8,
  group = "glpolygons",
  popup = NULL,
  label = NULL,
  layerId = NULL,
  src = FALSE,
  pane = "overlayPane",
  stroke = TRUE,
  popupOptions = NULL,
  labelOptions = NULL,
  contextMenu = NULL,
  ...
)
}
\arguments{
\item{map}{a map widget object created from \code{\link[leaflet:leaflet]{leaflet()}}}

\item{data}{sf/sp point/polygon/line data to add to the map.}

\item{color}{Object representing the color. Can be of class integer, character with
color names, HEX codes or random characters, factor, matrix, data.frame, list, json or formula.
See the examples or \link{makeColorMatrix} for more information.}

\item{opacity}{feature opacity. Numeric between 0 and 1.
Note: expect funny results if you set this to < 1.}

\item{group}{the name of the group the newly created layers should belong to
(for \code{\link[leaflet:clearGroup]{clearGroup()}} and \code{\link[leaflet:addLayersControl]{addLayersControl()}} purposes).
Human-friendly group names are permitted--they need not be short,
identifier-style names. Any number of layers and even different types of
layers (e.g., markers and polygons) can share the same group name.}

\item{popup}{Object representing the popup. Can be of type character with column names,
formula, logical, data.frame or matrix, Spatial, list or JSON. If the length does not
match the number of rows in the data, the popup vector is repeated to match the dimension.}

\item{label}{a character vector of the HTML content for the labels}

\item{weight}{line width/thickness in pixels for \code{addGlPolylines}.}

\item{layerId}{the layer id}

\item{src}{whether to pass data to the widget via file attachments.}

\item{pane}{A string which defines the pane of the layer. The default is \code{"overlayPane"}.}

\item{popupOptions}{A Vector of \code{\link[leaflet:popupOptions]{popupOptions()}} to provide popups}

\item{labelOptions}{A Vector of \code{\link[leaflet:labelOptions]{labelOptions()}} to provide label
options for each label. Default \code{NULL}}

\item{contextMenu}{a \code{\link[htmlwidgets]{JS}} function, that is passed to contextmenu.
See the example in \code{./inst/examples/contextmenu.R}}

\item{...}{Used to pass additional named arguments to \code{\link[yyjsonr]{write_json_str}} or
\code{\link[yyjsonr]{write_geojson_str}} & to pass additional arguments to the
underlying JavaScript functions. Typical use-cases include setting \code{'digits'} to
round the point coordinates or to pass a different \code{'fragmentShaderSource'} to
control the shape of the points. Use
\itemize{
   \item{\code{'point'} (default) to render circles with a thin black outline}
   \item{\code{'simpleCircle'} for circles without outline}
   \item{\code{'square'} for squares without outline}
}
Additional arguments could be \code{'sensitivity'}, \code{'sensitivityHover'} or
\code{'vertexShaderSource'}. See a full list at the
\href{https://github.com/robertleeplummerjr/Leaflet.glify}{Leaflet.glify}
repository.}

\item{fillColor}{fill color}

\item{fillOpacity}{fill opacity}

\item{radius}{point size in pixels.}

\item{stroke}{whether to draw stroke along the path (e.g., the borders of
polygons or circles)}
}
\description{
Leaflet.glify is a WebGL renderer plugin for leaflet. See
  \url{https://github.com/robertleeplummerjr/Leaflet.glify} for details
  and documentation.
}
\section{Functions}{
\itemize{
\item \code{addGlPolylines()}: Add Lines to a leaflet map using Leaflet.glify

\item \code{addGlPoints()}: Add Points to a leaflet map using Leaflet.glify

\item \code{addGlPolygons()}: Add Polygons to a leaflet map using Leaflet.glify

}}
\note{
MULTILINESTRINGs and MULTIPOLYGONs are currently not supported!
  Make sure you cast your data to LINESTRING or POLYGON first using:
  \itemize{
     \item{\code{sf::st_cast(data, "LINESTRING")}}
     \item{\code{sf::st_cast(data, "POLYGON")}}
  }
}
\section{Shiny Inputs}{

  The objects created with \code{leafgl} send input values to Shiny as the
  user interacts with them. These events follow the pattern
  \code{input$MAPID_glify_EVENTNAME}.
  The following events are available:

  \itemize{
    \item \strong{Click Events:}
      \code{input$MAPID_glify_click}
    \item \strong{Mouseover Events:}
      \code{input$MAPID_glify_mouseover}
    \item \strong{Mouseout Events:}
      \code{input$MAPID_glify_mouseout}
  }


  Each event returns a list containing:
  \itemize{
    \item \code{lat}: Latitude of the object or mouse cursor
    \item \code{lng}: Longitude of the object or mouse cursor
    \item \code{id}: The layerId, if any
    \item \code{group}: The group name of the object
    \item \code{data}: The properties of the feature
  }
}

\examples{
library(leaflet)
library(leafgl)
library(sf)

storms = st_as_sf(atlStorms2005)
cols = heat.colors(nrow(storms))

leaflet() \%>\%
  addProviderTiles(provider = providers$CartoDB.Positron) \%>\%
  addGlPolylines(data = storms, color = cols, popup = TRUE, opacity = 1)

\donttest{
library(leaflet)
library(leafgl)
library(sf)

n = 1e5
df1 = data.frame(id = 1:n,
                 x = rnorm(n, 10, 1),
                 y = rnorm(n, 49, 0.8))
pts = st_as_sf(df1, coords = c("x", "y"), crs = 4326)
cols = topo.colors(nrow(pts))

leaflet() \%>\%
  addProviderTiles(provider = providers$CartoDB.DarkMatter) \%>\%
  addGlPoints(data = pts, fillColor = cols, popup = TRUE)
}
library(leaflet)
library(leafgl)
library(sf)

gadm = st_as_sf(gadmCHE)
gadm = st_cast(gadm, "POLYGON")
cols = grey.colors(nrow(gadm))

leaflet() \%>\%
  addProviderTiles(provider = providers$CartoDB.DarkMatter) \%>\%
  addGlPolygons(data = gadm, color = cols, popup = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-color.R
\name{makeColorMatrix}
\alias{makeColorMatrix}
\title{makeColorMatrix}
\usage{
makeColorMatrix(x, data, palette, ...)
}
\arguments{
\item{x}{Object representing the color. Can be of class integer, numeric, Date,
POSIX*, character with color names or HEX codes, factor, matrix, data.frame,
list, json or formula.}

\item{data}{The dataset}

\item{palette}{Name of a color palette. If \code{colourvalues} is installed, it is
passed to \code{\link[colourvalues]{colour_values_rgb}}. To see all available
palettes, please use \code{\link[colourvalues]{colour_palettes}}.
If \code{colourvalues} is not installed, the palette is passed to
\code{\link[leaflet]{colorNumeric}}.}

\item{...}{Passed to \code{\link[colourvalues]{colour_palettes}} or
\code{\link[leaflet]{colorNumeric}}.}
}
\description{
Transform object to rgb color matrix
}
\examples{
{
## For Integer/Numeric/Factor
makeColorMatrix(23L)
makeColorMatrix(23)
makeColorMatrix(as.factor(23))

## For POSIXt / Date
makeColorMatrix(as.POSIXlt(Sys.time(), "America/New_York"), NULL)
makeColorMatrix(Sys.time(), NULL)
makeColorMatrix(Sys.Date(), NULL)

## For matrix/data.frame
makeColorMatrix(cbind(130,1,1), NULL)
makeColorMatrix(matrix(1:99, ncol = 3, byrow = TRUE), data.frame(x=c(1:33)))
makeColorMatrix(data.frame(matrix(1:99, ncol = 3, byrow = TRUE)), data.frame(x=c(1:33)))

## For characters
testdf <- data.frame(
  texts = LETTERS[1:10],
  vals = 1:10,
  vals1 = 11:20
)
makeColorMatrix("red", testdf)
makeColorMatrix("val", testdf)

## For formulaes
makeColorMatrix(~vals, testdf)
makeColorMatrix(~vals1, testdf)

## For JSON
makeColorMatrix(leafgl:::yyson_json_str(data.frame(r = 54, g = 186, b = 1)), NULL)

## For Lists
makeColorMatrix(list(1,2), data.frame(x=c(1,2)))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/leafgl_package.R
\docType{package}
\name{leafgl-package}
\alias{leafgl}
\alias{leafgl-package}
\title{leafgl: High-Performance 'WebGl' Rendering for Package 'leaflet'}
\description{
Provides bindings to the 'Leaflet.glify' JavaScript library which extends the 'leaflet' JavaScript library to render large data in the browser using 'WebGl'.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/r-spatial/leafgl}
  \item \url{https://r-spatial.github.io/leafgl/}
  \item Report bugs at \url{https://github.com/r-spatial/leafgl/issues}
}

}
\author{
\strong{Maintainer}: Tim Appelhans \email{tappelhans@tutamail.com} (\href{https://orcid.org/0000-0002-9824-2707}{ORCID}) [copyright holder]

Other contributors:
\itemize{
  \item Colin Fay \email{colin@thinkr.fr} (\href{https://orcid.org/0000-0001-7343-1846}{ORCID}) [contributor]
  \item Robert Plummer (Leaflet.glify plugin) [contributor]
  \item Kent Johnson \email{kjohnson@akoyabio.com} [contributor]
  \item Sebastian Gatscha \email{sebastian_gatscha@gmx.at} [contributor]
  \item Olivier Roy [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-color.R
\name{checkDim}
\alias{checkDim}
\title{checkDim}
\usage{
checkDim(x, data)
}
\arguments{
\item{x}{The color vector}

\item{data}{The dataset}
}
\description{
Check the length of the color vector. It must match the
  number of rows of the dataset.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/glify-remove-clear.R
\name{remove}
\alias{remove}
\alias{removeGlPoints}
\alias{removeGlPolylines}
\alias{removeGlPolygons}
\alias{clearGlLayers}
\alias{clearGlGroup}
\title{Remove Leaflet.Glify elements from a map}
\usage{
removeGlPoints(map, layerId)

removeGlPolylines(map, layerId)

removeGlPolygons(map, layerId)

clearGlLayers(map)

clearGlGroup(map, group)
}
\arguments{
\item{map}{a map widget object, possibly created from \code{\link[leaflet:leaflet]{leaflet()}}
but more likely from \code{\link[leaflet:leafletProxy]{leafletProxy()}}}

\item{layerId}{character vector; the layer id(s) of the item to remove}

\item{group}{the name of the group whose members should be removed}
}
\value{
the new `map` object
}
\description{
Remove one or more features from a map, identified by `layerId`;
or, clear all features of the given group.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/glify-shiny.R
\name{leafglOutput}
\alias{leafglOutput}
\alias{renderLeafgl}
\title{Use leafgl in shiny}
\usage{
leafglOutput(outputId, width = "100\%", height = 400)

renderLeafgl(expr, env = parent.frame(), quoted = TRUE)
}
\arguments{
\item{outputId}{output variable to read from}

\item{width, height}{the width and height of the map}

\item{expr}{An expression that generates an HTML widget}

\item{env}{The environment in which to evaluate expr.}

\item{quoted}{Is expr a quoted expression (with quote())?
This is useful if you want to save an expression in a variable.}
}
\value{
A UI for rendering leafgl

A server function for rendering leafgl
}
\description{
Use leafgl in shiny
}
\details{
See \link[leaflet:map-shiny]{leaflet::leafletOutput} for details.
  \code{renderLeafgl} is only exported for consistency. You can just as well
  use \link[leaflet:map-shiny]{leaflet::renderLeaflet} (see example).
  \code{leafglOutput} on the other hand is needed as it will attach all
  necessary dependencies.
}
\examples{
if (interactive()) {
library(shiny)
library(leaflet)
library(leafgl)
library(sf)

n = 1e4
df1 = data.frame(id = 1:n,
    x = rnorm(n, 10, 3),
    y = rnorm(n, 49, 1.8))
pts = st_as_sf(df1, coords = c("x", "y"), crs = 4326)

m = leaflet() \%>\%
 addProviderTiles(provider = providers$CartoDB.DarkMatter) \%>\%
 addGlPoints(data = pts, group = "pts") \%>\%
 setView(lng = 10.5, lat = 49.5, zoom = 6) \%>\%
 addLayersControl(overlayGroups = "pts")

ui <- fluidPage(
    leafglOutput("mymap")
)

server <- function(input, output, session) {
    output$mymap <- renderLeaflet(m)
}

shinyApp(ui, server)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-popup.R
\name{makePopup}
\alias{makePopup}
\title{makePopup}
\usage{
makePopup(x, data)
}
\arguments{
\item{x}{Object representing the popup}

\item{data}{The dataset}
}
\description{
Transform object to popup
}
